'''
    pypsa2bb - create input data for the esm backbone, run backbone and analyse results
    Copyright (C) 2020-2022 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from plotly.subplots import make_subplots
import plotly.graph_objects as go
import numpy as np
from results import PypsaResult, BackboneResult
import pandas as pd
from helpers import configure_logger
import sys
import json


# check if snakemake variable is present
if "snakemake" in locals():
    # this script has been called as part of a snakemake workflow
    # modify input and output paths accordingly
    logger = configure_logger(snakemake.rule, snakemake=snakemake)
    config = snakemake.config
    config["pypsa2bbPaths"]["pypsa_result_path"] = snakemake.input.pypsa
    config["pypsa2bbPaths"]["backbone_result_path"] = snakemake.input.backbone
    config["pypsa2bbPaths"]["figure_path"] = snakemake.output[0]
elif len(sys.argv) > 1:
    # this script has been called from the command line and paths were specified in command
    config["pypsa2bbPaths"]["pypsa_result_path"] = sys.argv[1]
    config["pypsa2bbPaths"]["backbone_result_path"] = sys.argv[2]
    config["pypsa2bbPaths"]["figure_path"] = sys.argv[3]
    logger = configure_logger(__file__, config=config)
else:
    # this script has been called from command line and paths were not specified.
    # Default paths from config.json will be used.
    logger = configure_logger(__file__, config=config)
    logger.warning(
        "comparison.py called without snakemake or args. All paths will be derived from config."
    )

# store current config for debugging
# json.dump(config, open(config_path, "w"))

# # Comparing results
# ## 1) getting results from pypsa

# %%
pp_results = PypsaResult(config["pypsa2bbPaths"]["pypsa_result_path"])
pypsa_country_dfs = pp_results.country_frames

# aggregated for all countries:
if len(pypsa_country_dfs.keys()) > 1:
    df_list = [pypsa_country_dfs[c] for c in pypsa_country_dfs.keys()]

    # concatenation yields long dataframe with repeating index
    all_countries = pd.concat(df_list).groupby("name").sum()
    pypsa_country_dfs["all"] = all_countries


# %%
# 2) get results from backbone
bb_results = BackboneResult(config["pypsa2bbPaths"]["backbone_result_path"])
bb_country_dfs = bb_results.get_country_frames()

# aggregated for all countries:
if len(bb_country_dfs.keys()) > 1:
    df_list = [bb_country_dfs[c] for c in bb_country_dfs.keys()]

    # concatenation yields long dataframe with repeating index
    all_countries = pd.concat(df_list).groupby("t").sum()
    bb_country_dfs["all"] = all_countries


na_counts = [test.isna().sum().sum() for test in pypsa_country_dfs.values()]
na_counts

# # Calculating differences

# check countries:
diff_countries = set(bb_country_dfs.keys()).symmetric_difference(
    pypsa_country_dfs.keys()
)
# the difference should be 0
if len(diff_countries) != 0:
    logger.warning(
        f"Backbone and Pypsa operated on different countries. {diff_countries} not present in both"
    )
    pass

# compare only common countries
compare_countries = list(set(bb_country_dfs.keys()) & set(pypsa_country_dfs.keys()))


def rmse(arr1, arr2):
    return np.sqrt(np.mean(np.square(arr1 - arr2)))


def nrmse(arr1, arr2):
    m = np.mean(arr1)
    return rmse(arr1, arr2) / m


# carriers:
comparison_results = {}
for country in compare_countries:
    pypsa_result = pypsa_country_dfs[country]
    bb_result = bb_country_dfs[country]
    bb_key_set = set(bb_result.keys())
    pp_key_set = set(pypsa_result.keys())
    diff_carriers = bb_key_set.symmetric_difference(pp_key_set)
    compare_carriers = list(set(bb_result.keys()) & set(pypsa_result.keys()))
    if len(diff_carriers) != 0:
        msg = (
            f'Found carriers differing carriers in "{country}".'
            f"\n\tbackbone : {bb_key_set}\n\tpypsa\t: {pp_key_set}:\n\tComparing {compare_carriers}"
        )
        logger.warning(msg)

    # dict for storing values
    carrier_comparison = {}
    for carrier in compare_carriers:

        pypsa_c = pypsa_result[carrier]
        bb_c = bb_result[carrier]
        err = nrmse(pypsa_c, bb_c)

        # add result of this carrier to country result
        carrier_comparison.update({carrier: err})

    # add country result to total results
    comparison_results.update({country: carrier_comparison})

# %%

bb_carriers = set(bb_results.carriers)
pypsa_carriers = set(pp_results.carriers)
# use carrier names from model with most carriers
# missing carriers will be filled with 0s for plotting
if len(bb_carriers) > len(pypsa_carriers):
    no_carriers = len(bb_carriers)
    carriers_for_plot = bb_carriers
else:
    no_carriers = len(pypsa_carriers)
    carriers_for_plot = pypsa_carriers


subplot_titles = sorted(carriers_for_plot) + ["aggregated"]


fig = make_subplots(
    rows=no_carriers + 1,
    cols=1,
    vertical_spacing=0.03,
    row_heights=[3] * (no_carriers + 1),
    shared_xaxes=True,
    row_titles=subplot_titles,
)
no_traces = (no_carriers + 1) * 2 * len(compare_countries)

btns = [
    {
        "label": "select country",
        "method": "update",
        "args": [{"visible": [False] * no_traces}],
    }
]
for i, c in enumerate(compare_countries):
    visibility = [False] * no_traces
    # turn visibility of carriers for current country true
    shift_length = (no_carriers + 1) * 2
    visibility[i * shift_length : (i + 1) * shift_length] = [True] * shift_length
    btn = {"label": c, "method": "update", "args": [{"visible": visibility}]}
    btns.append(btn)
    for j, carrier in enumerate(subplot_titles[:-1]):
        t = pypsa_country_dfs[c].index

        # use this flag to signal missing carriers
        pp_flag, bb_flag = "", ""
        if carrier in pypsa_country_dfs[c].columns:
            pp_c = pypsa_country_dfs[c][carrier]
        else:
            pp_c = [0] * len(t)
            pp_flag = "-NA"
        if carrier in bb_country_dfs[c].columns:
            bb_c = bb_country_dfs[c][carrier]
        else:
            bb_c = [0] * len(t)
            bb_flag = "-NA"

        fig.add_trace(
            go.Scatter(
                x=t,
                y=pp_c,
                name=f"pp-{c}-{carrier}{pp_flag}",
                visible=False,
                line={"color": "red"},
            ),
            row=j + 1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=t,
                y=bb_c,
                name=f"bb-{c}-{carrier}{bb_flag}",
                visible=False,
                line={"color": "blue"},
            ),
            row=j + 1,
            col=1,
        )
        pass
    pp_agg = pypsa_country_dfs[c].sum(axis=1)
    bb_agg = bb_country_dfs[c].sum(axis=1)
    fig.add_trace(
        go.Scatter(
            x=t, y=pp_agg, name=f"pp-{c}-agg", visible=False, line={"color": "red"}
        ),
        row=j + 2,
        col=1,
    )
    fig.add_trace(
        go.Scatter(
            x=t, y=bb_agg, name=f"bb-{c}-agg", visible=False, line={"color": "blue"}
        ),
        row=j + 2,
        col=1,
    )
    pass


fig.update_layout(
    {
        "updatemenus": [
            go.layout.Updatemenu(
                buttons=btns,
                direction="down",
                pad={"r": 5, "t": 8},
                showactive=True,
                xanchor="right",
                x=1.3,
                yanchor="top",
                y=1.10,
            )
        ],
        "title_text": "Production by carrier",
        "height": 1000,
        "width": 1300,
    }
)


fig.write_html(config["pypsa2bbPaths"]["figure_path"], auto_open=False)
