'''
    pypsa2bb - create input data for the esm backbone, run backbone and analyse results
    Copyright (C) 2020-2022 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import pandas as pd
from pathlib import Path
import constants as const
import os
import xarray as xr
import pypsa
from scipy.sparse.csgraph import dijkstra
import sys
sys.path.append("pypsa-eur/scripts")
from simplify_network import simplify_network_to_380

fuels = ["gas", "coal", "lignite",  "oil", "nuclear", "biomass", "geothermal", "other"]
renewableTechs = ["wind", "solar", "ror", "biomass", "geothermal", "nuclear", "hydro discharge"]


class NC2XLS_Converter:
    def __init__(self, config):
        self.config = config

        self.costs = self._load_costs(self.config["pypsa2bbPaths"]["costs_filename"])
        if self.config["updateCosts"]["useUpdatedCosts"]:
            self.updatedCosts = self._load_costs("resources/updatedCosts.csv")
            self.costs = self._combine_costs_and_updatedCosts()
        self.pypsaDs = xr.load_dataset(
            self.config["pypsa2bbPaths"]["pypsa_model_path"]
        )
        self.numberOfBuses = const.num_buses(config)
        self.temporalResolution = const.temporal_resolution(config)

        self.pypsa_path = Path(config["pypsa2bbPaths"]["pypsafolder_path"])
        # set certain attributes that are required for multiple sheets
        self._set_timesteps()
        self._set_storage_nodes()
        self._set_units()
        self._set_techs()
        pass

    def _load_costs(self, filename):
        filename_costs = filename
        costs_file = pd.read_csv(filename_costs, index_col=list(range(2))).sort_index()
        # correct units to MW and EUR
        costs_file.loc[costs_file.unit.str.contains("/kW"), "value"] *= 1e3
        costs_file.loc[costs_file.unit.str.contains("USD"), "value"] *= self.config["USD_to_EUR"]
        costs = costs_file["value"].unstack(level=1).groupby("technology").sum(min_count=1)
        costs = costs.fillna(
            {
                "CO2 intensity": 0,
                "FOM": 0,
                "VOM": 0,
                "discount rate": self.config["costs"]["fill_values"]["discount rate"],
                "efficiency": 1,
                "fuel": 0,
                "investment": 0,
                "lifetime": 25,
            }
        )
        costs.loc["none"] = 0

        return costs

    def _combine_costs_and_updatedCosts(self):
        techsUpdated = [tech for tech in self.updatedCosts.index]
        techsNotUpdated = [tech for tech in self.costs.index if not tech in self.updatedCosts.index]
        costs_combined = pd.concat([self.updatedCosts.loc[techsUpdated],
                                    self.costs.loc[techsNotUpdated]])
        techsNotUpdated = set([tech for tech in techsNotUpdated if not tech == "none"])
        if len(techsNotUpdated) > 0:
            for element in techsNotUpdated:
                print(f"no updated costs found for {element}. Using default pypsa-eur value.")

        return costs_combined

    def _set_units(self):
        unit = list(self.pypsaDs["generators_i"].to_dataframe()["generators_i"])
        unit = unit + [element + " charge unit" for element in self.storage_nodes] \
            + [element + " discharge unit" for element in self.storage_nodes]
        
        if self.config["addAdvancedStorages"]:
            unit = unit + [element + " storage unit" for element in self.storage_nodes  
                           if ((not "hydro" in element) and (not "PHS" in element))]

        self.units = unit

    def _set_storage_nodes(self):
        if self.config["addAdvancedStorages"]:
            nodeList = []
            for storageType in ["VRFB", "LIB", "ACAES", "H2"]:
                for node in self.node:
                    nodeList.append(f"{node} {storageType}")
            for storageNode in self.pypsaDs["storage_units_i"].to_dataframe()["storage_units_i"].tolist():
                if "hydro" in storageNode or "PHS" in storageNode: # keep hydro nodes
                    nodeList.append(storageNode)
            self.storage_nodes = nodeList
        else:
            self.storage_nodes = self.pypsaDs["storage_units_i"].to_dataframe()["storage_units_i"].tolist()

    def _set_timesteps(self):
        # timeframes
        snapshots = self.pypsaDs["snapshots"].to_dataframe()
        self.timesteps = [f"t{i:06d}" for i in range(1, len(snapshots)+1)]
        self.datalength = len(self.pypsaDs["snapshots"].to_dataframe())

    def _set_techs(self):
        techs = [
            name.split()[2] if not "charge" in name and not "storage" in name 
            else f"{name.split()[2]} {name.split()[3]}" for name in self.units]
        techs = ["offwind" if "offwind" in tech else tech for tech in techs]
        if self.config["addAdvancedStorages"]:
            techs = ["H2-GT" if "H2 discharge" in tech else tech for tech in techs]
            techs = ["PEM" if "H2 charge" in tech else tech for tech in techs]   
        else:
            techs = ["fuel cell" if "H2 discharge" in tech else tech for tech in techs]
            techs = ["electrolysis" if "H2 charge" in tech else tech for tech in techs]
        techs = ["charge_battery inverter" if "battery charge" in tech else tech for tech in techs]
        techs = ["charge_hydro" if "hydro charge" in tech else tech for tech in techs]
        techs = ["charge_PHS" if "PHS charge" in tech else tech for tech in techs]
        techs = ["charge_LIB" if "LIB charge" in tech else tech for tech in techs]
        techs = ["charge_VRFB" if "VRFB charge" in tech else tech for tech in techs]
        techs = ["charge_ACAES" if "ACAES charge" in tech else tech for tech in techs]
        techs = ["battery inverter" if "battery discharge" in tech else tech for tech in techs]
        techs = ["hydro" if "hydro discharge" in tech else tech for tech in techs]
        techs = ["PHS" if "PHS discharge" in tech else tech for tech in techs]
        techs = ["LIB" if "LIB discharge" in tech else tech for tech in techs]
        techs = ["VRFB" if "VRFB discharge" in tech else tech for tech in techs]
        techs = ["ACAES" if "ACAES discharge" in tech else tech for tech in techs]     
        techs = ["LIB-storage" if "LIB storage" in tech else tech for tech in techs]
        techs = ["VRFB-storage" if "VRFB storage" in tech else tech for tech in techs]
        techs = ["ACAES-storage" if "ACAES storage" in tech else tech for tech in techs]
        techs = ["H2-storage" if "H2 storage" in tech else tech for tech in techs]
        self.techs = techs

    def get_group(self):
        # add all necessary groups to group sheet
        group_r = list()
        for element in self.storage_nodes:
            group_r.append(element + " capgroup1")
            group_r.append(element + " capgroup2")
        if self.config["constraints"]["energySharePerCountry"]:
            for element in self.config["constraints"]["countriesEnergyShare"]:
                group_r.append("energyShareGroup " + element)
        if self.config["constraints"]["energyShareTotal"]:
            group_r.append("energyShareGroup total")
        if self.config["constraints"]["CO2Tax"]:
            group_r.append("CO2TaxGroup")
        if self.config["constraints"]["emissionCap"]:
            group_r.append("emissionCapGroup")
        if self.config["constraints"]["addBiomassConstraint"]:
            countries = list(set([node[0:2] for node in self.node]))
            for country in countries:
                group_r.append(f"{country} biomass cap group")
        return pd.DataFrame(data=group_r, columns=["group"])

    def get_ugroup(self):
        # define ugroup for forcing the same amount of investment in charge and discharge units
        constrainedNodes = [node for node in self.storage_nodes if not "H2" in node]
        ugroup_group_r = [f'{element} capgroup1' for element in constrainedNodes]*2 \
            + [f'{element} capgroup2' for element in constrainedNodes]*2
        ugroup_unit_r = ([f'{element} charge unit' for element in constrainedNodes]
                         + [f'{element} discharge unit' for element in constrainedNodes])*2
        ugroup_group_df = pd.DataFrame(data=ugroup_group_r, columns=["group"])
        ugroup_unit_df = pd.DataFrame(data=ugroup_unit_r, columns=["unit"])
        return pd.concat([ugroup_unit_df, ugroup_group_df], axis=1, sort=False)

    def get_gngroup(self):
        # add nodes and grids to gngroup for energyShare constraints, CO2 taxes and emission cap constraints
        if (
            self.config["constraints"]["energySharePerCountry"]
            or self.config["constraints"]["energyShareTotal"]
            or self.config["constraints"]["CO2Tax"]
            or self.config["constraints"]["emissionCap"]
        ):
            gngroup_grid_r = list()
            gngroup_node_r = list()
            gngroup_group_r = list()
            p_gn = self.get_p_gn()

            if self.config["constraints"]["energySharePerCountry"]:
                gngroup_grid_r = [grid for grid, node in zip(
                    p_gn["grid"], p_gn["node"]) if grid == "elec" and node[0:2] in self.config["constraints"]["countriesEnergyShare"]]
                gngroup_node_r = [node for grid, node in zip(
                    p_gn["grid"], p_gn["node"]) if grid == "elec" and node[0:2] in self.config["constraints"]["countriesEnergyShare"]]
                gngroup_group_r = ["energyShareGroup " + node[0:2] for grid, node in zip(p_gn["grid"], p_gn["node"])
                                   if grid == "elec" and node[0:2] in self.config["constraints"]["countriesEnergyShare"]]

            if self.config["constraints"]["energyShareTotal"]:
                gngroup_grid_r = gngroup_grid_r + [grid for grid in p_gn["grid"] if grid == "elec"]
                gngroup_node_r = gngroup_node_r + [node for grid,
                                                   node in zip(p_gn["grid"], p_gn["node"]) if grid == "elec"]
                gngroup_group_r = gngroup_group_r + \
                    ["energyShareGroup total" for grid in p_gn["grid"] if grid == "elec"]

            if self.config["constraints"]["CO2Tax"]:
                fuelEmission = self.get_p_nemission()
                gngroup_grid_r = gngroup_grid_r + ["fuel"]*len(fuelEmission["node"])
                gngroup_node_r = gngroup_node_r + fuelEmission["node"].tolist()
                gngroup_group_r = gngroup_group_r + ["CO2TaxGroup"]*len(fuelEmission["node"])

            if self.config["constraints"]["emissionCap"]:
                fuelEmission = self.get_p_nemission()
                gngroup_grid_r = gngroup_grid_r + ["fuel"]*len(fuelEmission["node"])
                gngroup_node_r = gngroup_node_r + fuelEmission["node"].tolist()
                gngroup_group_r = gngroup_group_r + ["emissionCapGroup"]*len(fuelEmission["node"])

            gngroup_grid_df = pd.DataFrame(data=gngroup_grid_r, columns=["grid"])
            gngroup_node_df = pd.DataFrame(data=gngroup_node_r, columns=["node"])
            gngroup_group_df = pd.DataFrame(data=gngroup_group_r, columns=["group"])
            return pd.concat(
                [gngroup_grid_df, gngroup_node_df, gngroup_group_df], axis=1, sort=False
            )
        else:
            return pd.DataFrame([], columns=["grid", "node", "group"])

    def get_gnugroup(self):
        # add grids, nodes and unit to gnugroup for energyShare constraints
        p_gnu_io = self.get_p_gnu_io()
        gnugroup_df_final = pd.DataFrame(data = [], columns = ["grid", "node", "unit", "group"])

        # add renewable technologies to groups to enforce energy share constraints
        if self.config["constraints"]["energySharePerCountry"]:
            gnugroup_df = p_gnu_io.loc[p_gnu_io["grid"] == "elec"]
            gnugroup_df = pd.concat([gnugroup_df.loc[gnugroup_df["node"].str.contains(country)] for country in
                                     self.config["constraints"]["countriesEnergyShare"]])
            gnugroup_df = pd.concat([gnugroup_df.loc[gnugroup_df["unit"].str.contains(tech)]
                                    for tech in renewableTechs])
            gnugroup_df["group"] = [f"energyShareGroup {country[0:2]}" for country in gnugroup_df["node"]]
            gnugroup_df = gnugroup_df[["grid", "node", "unit", "group"]]
            gnugroup_df_final = pd.concat([gnugroup_df_final, gnugroup_df])

        if self.config["constraints"]["energyShareTotal"]:
            gnugroup_df = p_gnu_io.loc[p_gnu_io["grid"] == "elec"]
            gnugroup_df = pd.concat([gnugroup_df.loc[gnugroup_df["unit"].str.contains(tech)]
                                    for tech in renewableTechs])
            gnugroup_df["group"] = "energyShareGroup total"
            gnugroup_df = gnugroup_df[["grid", "node", "unit", "group"]]
            gnugroup_df_final = pd.concat([gnugroup_df_final, gnugroup_df])

        if self.config["constraints"]["addBiomassConstraint"]:
            gnugroup_df = p_gnu_io.loc[(p_gnu_io["grid"] == "fuel") & p_gnu_io["unit"].str.contains("biomass")]
            groupList = [f"{unit[0:2]} biomass cap group" for unit in gnugroup_df["unit"]]
            gnugroup_df["group"] = groupList
            gnugroup_df = gnugroup_df[["grid", "node", "unit", "group"]]
            gnugroup_df_final = pd.concat([gnugroup_df_final, gnugroup_df])

        return gnugroup_df_final

    def get_sgroup(self):
        # add sample s000 to groups for energyShare constraints and emission cap constraint
        if (
            self.config["constraints"]["energySharePerCountry"]
            or self.config["constraints"]["energyShareTotal"]
            or self.config["constraints"]["emissionCap"]
            or self.config["constraints"]["addBiomassConstraint"]
        ):
            s_group = self.get_group()
            s_group = s_group.loc[s_group["group"].str.contains("Share") | s_group["group"].str.contains("Cap") 
                                  | s_group["group"].str.contains("biomass")]
            s_group["sample"] = "s000"
            return s_group[["sample", "group"]]
        else:
            return pd.DataFrame(data=[], columns=["sample", "group"])

    def get_p_groupPolicy(self):
        # add energyShare constraint values to p_groupPolicy
        if (
            self.config["constraints"]["energySharePerCountry"]
            or self.config["constraints"]["energyShareTotal"]
            or self.config["constraints"]["addBiomassConstraint"]
        ):
            pgp_df_1 = pd.DataFrame()
            pgp_df_2 = pd.DataFrame()
            pgp_df_3 = pd.DataFrame()
            if self.config["constraints"]["energySharePerCountry"]:
                pgp_df_1["group"] = [
                    f"energyShareGroup {element}" for element in self.config["constraints"]["countriesEnergyShare"]]
                pgp_df_1["policy"] = "energyShareMin"
                pgp_df_1["value"] = self.config["constraints"]["energySharePerCountryValue"]
            if self.config["constraints"]["energyShareTotal"]:
                pgp_df_2["group"] = ["energyShareGroup total"]
                pgp_df_2["policy"] = ["energyShareMin"]
                pgp_df_2["value"] = [self.config["constraints"]["energyShareTotalValue"]]
            if self.config["constraints"]["addBiomassConstraint"]:
                group = self.get_group()
                pgp_df_3["group"] = group.loc[group["group"].str.contains("biomass")]["group"]
                pgp_df_3["policy"] = "energyMax"
                potential = pd.read_csv("resources/biomassPotential.csv", sep=";", index_col=0, header=[1,2])
                potentialList = []
                for group in pgp_df_3["group"]:
                    country = group[0:2]
                    if country in potential.index:
                        potentialList.append(potential.loc[country, self.config["constraints"]["biomassConstraintScenario"][0]]
                                         [str(self.config["constraints"]["biomassConstraintScenario"][1])])
                    else:
                        potentialList.append(0)
                pgp_df_3["value"] = potentialList
            return pd.concat(
                [pgp_df_1, pgp_df_2, pgp_df_3], sort=False
            )
        else:
            return pd.DataFrame(data=[], columns=["group", "param_policy"])

    def get_p_groupPolicyUnit(self):
        # set up constraints for forcing the same amount of investment in charge and discharge units for storages
        pgp_df = pd.DataFrame()
        constrainedNodes = [node for node in self.storage_nodes if not "H2" in node]
        pgp_df["group"] = [f'{element} capgroup1' for element in constrainedNodes]*2 \
            + [f'{element} capgroup2' for element in constrainedNodes]*2
        pgp_df["param_policy"] = "constrainedCapMultiplier"
        pgp_df["unit"] = ([f'{element} charge unit' for element in constrainedNodes]
                          + [f'{element} discharge unit' for element in constrainedNodes])*2
        pgp_df["value"] = [1]*len(constrainedNodes)+[-1]*len(constrainedNodes) + \
            [-1]*len(constrainedNodes)+[1]*len(constrainedNodes)
        return pgp_df

    def get_p_groupPolicyEmission(self):
        # add values for emission caps
        if (self.config["constraints"]["emissionCap"]):
            pgpEmission = pd.DataFrame()
            pgpEmission["group"] = ["emissionCapGroup"]
            pgpEmission["policy"] = ["emissionCap"]
            pgpEmission["emission"] = ["CO2"]
            pgpEmission["value"] = [self.config["constraints"]["emissionCapValue"]]
            return pgpEmission

        else:

            return pd.DataFrame(
                [], columns=["group", "param_policy", "emission", "value"]
            )

    def get_ts_emissionPriceChange(self):
        # set values for emission prices
        if (self.config["constraints"]["CO2Tax"]):
            return pd.DataFrame(
                ["CO2", "CO2TaxGroup", self.config["constraints"]["CO2TaxValue"]], index=["emission", "group", "t000000"]
            ).T
        else:
            return pd.DataFrame(
                [], columns=["emission", "group", "t000000"]
            )

    def get_availability_limits(self):
        # set availability for all units to first timestep
        availability_df = pd.DataFrame()
        availability_df["unit"] = self.units
        availability_df["timesteps"] = "t000001"
        availability_df["becomeAvailable"] = 1
        return availability_df

    def get_ts_influx_and_ts_node(self):
        # add demand and hydro inflows to ts_influx
        # starting with demand
        ts_influx_df = pd.DataFrame()
        ts_influx_df["grid"] = ["elec"] * len(self.node)
        ts_influx_df["node"] = self.node.tolist()
        ts_influx_df["forecast index"] = "f00"
        # multiply demand with -1 (must be negative in backbone) and divide by step length (if > 1, backbone and pypsa need different input)
        ts_influx_df = pd.concat([ts_influx_df, pd.DataFrame(data=(self.pypsaDs["loads_t_p_set"] * (-1) / float(self.temporalResolution)).T,
                                                             columns=self.timesteps)], axis=1)

        # scale total load if activated in config
        if self.config["updateCosts"]["scaleLoadTotal"]:
            ts_influx_df = ts_influx_df.set_index(["grid", "node", "forecast index"])
            ts_influx_df = ts_influx_df * (
                1 + self.config["updateCosts"]["scaleLoadTotalFactor"]
            )
            ts_influx_df.reset_index(inplace=True)

        # scale load per country (reverse the total load scaling for these countries, if also activated)
        if self.config["updateCosts"]["scaleLoadPerCountry"]:
            # calculate total Balkan load, because Balkan demand is grouped in Pietzcker et all
            if "Balkan" in self.config["updateCosts"]["scaleLoadPerCountryCountries"]:
                totalLoadBalkan = 0
                for country2 in ["AL", "BA", "ME", "MK", "RS"]:
                    totalLoadBalkan += - \
                        ts_influx_df.loc[ts_influx_df["node"].str.contains(country2)].sum(numeric_only=True).sum()
            ts_influx_df = ts_influx_df.set_index(["grid", "node", "forecast index"])

            for country, newLoad in zip(self.config["updateCosts"]["scaleLoadPerCountryCountries"],
                                        self.config["updateCosts"]["scaleLoadPerCountryNewLoad"]):
                if not self.config["updateCosts"]["scaleLoadTotal"]:
                    totalLoadFactor = 1
                else:
                    totalLoadFactor = self.config["updateCosts"]["scaleLoadTotalFactor"]

                # if new load is only given for complete Balkan area
                if country == "Balkan":
                    factor = self.config["updateCosts"]["scaleLoadPerCountryNewLoad"][self.config["updateCosts"]["scaleLoadPerCountryCountries"].index("Balkan")] \
                        / totalLoadBalkan*10**6
                    for country2 in ["AL", "BA", "ME", "MK", "RS"]:
                        ts_influx_df[ts_influx_df.index.get_level_values("node").str.contains(country2)] = \
                            ts_influx_df[ts_influx_df.index.get_level_values("node").str.contains(country2)] \
                            * factor/totalLoadFactor  # divide by total loadFactor to reverse total scaling
                else:
                    factor = -newLoad /\
                        ts_influx_df.loc[ts_influx_df.index.get_level_values(
                            "node").str.contains(country)].sum().sum()*10**6

                    ts_influx_df[ts_influx_df.index.get_level_values("node").str.contains(country)] = \
                        ts_influx_df[ts_influx_df.index.get_level_values("node").str.contains(country)] \
                        * factor/totalLoadFactor  # divide by total loadFactor to reverse total scaling
            ts_influx_df.reset_index(inplace=True)

        # add hydro inflows, if available
        storage_influx_df = pd.DataFrame()

        if "storage_units_t_inflow" in self.pypsaDs:
            storage_influx_df["grid"] = ["storage"]*len(self.pypsaDs["storage_units_t_inflow_i"])
            storage_influx_df["node"] = self.pypsaDs["storage_units_t_inflow_i"]
            storage_influx_df["forecast index"] = "f00"
            storage_influx_df = pd.concat([storage_influx_df, pd.DataFrame(
                data=self.pypsaDs["storage_units_t_inflow"].T, columns=self.timesteps)], axis=1)
            ts_influx_df = pd.concat([ts_influx_df, storage_influx_df])

        storage_influx_df["param_gnBoundaryTypes"] = "maxSpill"
        storage_influx_df = storage_influx_df[["grid", "node", "param_gnBoundaryTypes", "forecast index"]
                                              + storage_influx_df.columns.difference(["grid", "node", "param_gnBoundaryTypes", "forecast index"]).to_list()]

        return [ts_influx_df, storage_influx_df]

    def get_ts_cf(self):
        # add capacity factor time series
        ts_cf_df = self.pypsaDs["generators_t_p_max_pu"].to_dataframe().reset_index()\
            .pivot(columns="snapshots", index="generators_t_p_max_pu_i")
        ts_cf_df.columns = self.timesteps
        ts_cf_df["forecast index"] = "f00"
        ts_cf_df["node"] = ts_cf_df.index.str.rsplit(pat=" ", n=1, expand=True).get_level_values(0)
        ts_cf_df["flow"] = ts_cf_df.index.str.rsplit(pat=" ", n=1, expand=True).get_level_values(1)
        return ts_cf_df.reset_index(names="unit")[["flow", "node", "forecast index"] + self.timesteps]

    def get_ts_PriceChange(self):
        # write fuel price to ts_priceChange
        ts_priceChange_df = pd.DataFrame()
        ts_priceChange_df = self.costs.loc[self.fuelNodes]["fuel"].reset_index()
        return ts_priceChange_df[["technology", "fuel"]].rename(columns={"fuel": "t000000"})

    def get_p_unit(self):
        # remove _charge from techs
        techs = [tech.split("_")[1] if "charge_" in tech else tech for tech in self.techs]
        effsList = [self.costs.at[tech, "efficiency"] for tech in techs]
        p_unit_df = pd.DataFrame(data=effsList, index=self.unit["unit"], columns=["eff00"])

        # remove eff values for flow units to avoid backbone warnings
        p_unit_df = p_unit_df.astype("object")
        for unit in self.pypsaDs["generators_t_p_max_pu_i"].to_dataframe()["generators_t_p_max_pu_i"]:
            p_unit_df.loc[unit, "eff00"] = ""

        # set availability, accounting for deactivated power plants in file PPperCountry if activated
        p_unit_df["availability"] = 1
        if os.path.isfile("resources/PPperCountry.csv") and self.config["choosePPsPerCountry"]:
            excluded_PP = self.get_excluded_PP()
            for tech, country in excluded_PP:
                p_unit_df.loc[p_unit_df.index.str.contains(
                    tech) & p_unit_df.index.str.contains(country), "availability"] = 0
        # deactivate hydro charge units
        for unit in self.units:
            if "hydro charge" in unit:
                p_unit_df.loc[unit, "availability"] = 0

        # determine which units are extendable
        if self.config["addAdvancedStorages"]:            
            extendableChargeUnits = [unit for unit in self.units if " charge" in unit]
            extendable = [True if ("VRFB" in unit or "LIB" in unit or "ACAES" in unit or "H2" in unit) 
                          else False for unit in extendableChargeUnits]
            extendableCharge = pd.DataFrame([], index = extendableChargeUnits)
            extendableCharge["extendable"] = extendable
            extendableDischargeUnits = [unit for unit in self.units if " discharge" in unit]
            extendable = [True if ("VRFB" in unit or "LIB" in unit or "ACAES" in unit or "H2" in unit) 
                          else False for unit in extendableDischargeUnits]
            extendableDischarge = pd.DataFrame([], index = extendableDischargeUnits)
            extendableDischarge["extendable"] = extendable
        else:
            extendableCharge = self.pypsaDs["storage_units_p_nom_extendable"].to_dataframe()\
                .rename({"storage_units_p_nom_extendable": "extendable"}, axis=1)
            extendableCharge.index = [f"{name} charge unit" for name in extendableCharge.index]
            extendableDischarge = self.pypsaDs["storage_units_p_nom_extendable"].to_dataframe()\
                .rename({"storage_units_p_nom_extendable": "extendable"}, axis=1)
            extendableDischarge.index = [f"{name} discharge unit" for name in extendableDischarge.index]
        extendableGen = self.pypsaDs["generators_p_nom_extendable"].to_dataframe()\
            .rename({"generators_p_nom_extendable": "extendable"}, axis=1)
        extendable = pd.concat([extendableCharge, extendableDischarge, extendableGen])
        p_unit_df = pd.concat([p_unit_df, extendable], axis=1)
        p_unit_df.loc[p_unit_df.index.str.contains("invest"), "extendable"] = True

        for technology in ["CCGT", "OCGT", "nuclear", "biomass"]:
            if self.config["investOptionsBackbone"][technology]["investOption"]:
                if self.config["investOptionsBackbone"][technology]["countries"] == "all":
                    p_unit_df.loc[p_unit_df.index.str.contains(technology), "extendable"] = True
                else:
                    for investCountry in self.config["investOptionsBackbone"][technology][
                        "countries"
                    ]:
                        p_unit_df.loc[p_unit_df.index.str.contains(technology) 
                                      & p_unit_df.index.str.contains(investCountry), "extendable"] = True

        # get maximum capacity for all extendable units
        # maxUnitCount being inf for all extendable units without given maxCapacity
        capacity_max_r = self.pypsaDs["generators_p_nom_max"].to_dataframe()\
            .rename({"generators_p_nom_max": "maxUnitCount"}, axis=1).round(0)
        p_unit_df = pd.concat([p_unit_df, capacity_max_r], axis=1).infer_objects(copy=False).fillna("inf")
        p_unit_df.loc[p_unit_df["extendable"] == False, "maxUnitCount"] = 0

        if self.config["completeGreenField"]:       
            # get exitisting capacites for hydro/ror/PHS and set them as maximum invest
            maxCapRor = self.pypsaDs["generators_p_nom"].to_dataframe().rename(
                {"generators_p_nom": "maxUnitCount"}, axis=1)
            maxCapRor = maxCapRor.loc[maxCapRor.index.str.contains("ror")]

            p_unit_df = maxCapRor.combine_first(p_unit_df)

            # add storage capacities (hydro/PHS)
            if "storage_units_p_nom" in self.pypsaDs:
                maxCapHydroPHS = self.pypsaDs["storage_units_p_nom"].to_dataframe()\
                    .rename(columns={"storage_units_p_nom": "maxUnitCount"})
                maxCapHydroPHS.index = maxCapHydroPHS.index + " discharge unit"
                # remove units which are not in p_gnu_io_elec
                maxCapHydroPHS = maxCapHydroPHS.loc[maxCapHydroPHS.index.str.contains("hydro") 
                                                              | maxCapHydroPHS.index.str.contains("PHS")] 
                p_unit_df = maxCapHydroPHS.combine_first(p_unit_df)

            # set maximum capacity to infinity for all other units
            p_unit_df.loc[p_unit_df["maxUnitCount"] == 0, "maxUnitCount"] = "inf"                

        return p_unit_df.drop(columns=["extendable"]).reset_index(names="unit")

    def get_excluded_PP(self):
        chart = pd.read_csv("resources/PPperCountry.csv", delimiter=";", index_col=0)
        excluded_PP = list()
        for tech in chart.columns:
            for country in chart[tech].index:
                if chart.loc[country, tech] == 0:
                    excluded_PP.append([tech, country])
        return excluded_PP

    def get_p_nemission(self):
        unit_emission_df = self.pypsaDs["carriers_co2_emissions"].to_dataframe()\
            .rename({"carriers_co2_emissions": "CO2"}, axis=1)  # in tCO2/MWh
        unit_emission_df = unit_emission_df.loc[unit_emission_df["CO2"]
                                                > 0].drop("OCGT", errors="ignore").rename({"CCGT": "gas"})
        return unit_emission_df.reset_index(names="node")

    def get_flowunit(self):
        # assign flows to units
        flowunit_df = self.pypsaDs["generators_t_p_max_pu_i"].to_dataframe()

        # this creates e.g. "DE0 0", "onwind" from "DE0 0 onwind"
        flowunit_df["flow"] = flowunit_df.index.str.rsplit(pat=" ", n=1, expand=True).get_level_values(1)
        return flowunit_df.reset_index(drop=True).rename({"generators_t_p_max_pu_i": "unit"}, axis=1)[["flow", "unit"]]

    def get_p_gnu_io(self):
        # set charge units to none to avoid writing data to that
        techs = ["none" if "charge_" in tech else tech for tech in self.techs]
        # remove storage controlling units from electric grid
        techs = [tech for tech in techs if not "storage" in tech]
        elecUnits = [unit for unit in self.units if not "storage" in unit]
        p_gnu_io_elec = self.costs.loc[techs]
        p_gnu_io_elec = p_gnu_io_elec.reset_index(names="techs")
        p_gnu_io_elec.index = elecUnits

        # part of p_gnu_io where grid = elec
        p_gnu_io_elec = p_gnu_io_elec.rename(columns={"FOM": "fomCosts", "VOM": "vomCosts",
                                                      "investment": "invCosts"})
        # add marginal costs defined in pypsa config
        for key in self.config["costs"]["marginal_cost"].keys():
            p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains(
                key), "vomCosts"] = self.config["costs"]["marginal_cost"][key]
        p_gnu_io_elec["annuityFactor"] = p_gnu_io_elec["discount rate"] / (1 - (1 + p_gnu_io_elec["discount rate"]) **
                                                                           (-p_gnu_io_elec["lifetime"]))
        p_gnu_io_elec.drop(columns=["CO2 intensity", "fuel", "efficiency",
                           "discount rate", "lifetime"], inplace=True)

        # set conversionCoeff to 1 for all units except for flow units
        p_gnu_io_elec["conversionCoeff"] = 1
        flowunit_df = self.pypsaDs["generators_t_p_max_pu_i"].to_dataframe()
        p_gnu_io_elec.loc[flowunit_df.index, "conversionCoeff"] = 0

        # set capacities
        capacity_df = self.pypsaDs["generators_p_nom"].to_dataframe().rename(
            {"generators_p_nom": "capacity"}, axis=1)
        p_gnu_io_elec = p_gnu_io_elec.combine_first(capacity_df)

        # add storage capacities (hydro/PHS)
        if "storage_units_p_nom" in self.pypsaDs and not self.config["completeGreenField"]:
            capacity_storage_df = self.pypsaDs["storage_units_p_nom"].to_dataframe()\
                .rename(columns={"storage_units_p_nom": "capacity"})
            capacity_storage_df.index = capacity_storage_df.index + " discharge unit"
            # remove units which are not in p_gnu_io_elec
            capacity_storage_df = capacity_storage_df.loc[capacity_storage_df.index.isin(p_gnu_io_elec.index)]
            p_gnu_io_elec["capacity"] = p_gnu_io_elec["capacity"].combine_first(capacity_storage_df["capacity"])
        
        # remove all exisiting capacities for complete green field modelling
        if self.config["completeGreenField"]:
            p_gnu_io_elec["capacity"] = 0

        # fill unknown capacities with eps (0 won't be read by GAMS)
        p_gnu_io_elec = p_gnu_io_elec.astype('object')
        p_gnu_io_elec["capacity"] = p_gnu_io_elec["capacity"].fillna("eps").replace({0: "eps"})

        # all units get unitsize of 1
        p_gnu_io_elec["unitSize"] = 1

        # all units output into the electric grid except for storage charge units
        p_gnu_io_elec["input output"] = "output"
        p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains(" charge"), "input output"] = "input"
        # storage unit are input, should not have capacity
        p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains(" charge"), "capacity"] = 0

        # add node and grids
        p_gnu_io_elec["node"] = [f"{name.split(' ')[0]} {name.split(' ')[1]}" for name in p_gnu_io_elec.index]
        p_gnu_io_elec["grid"] = "elec"

        # multiply fom (in percent) with investment costs to get absolute fom costs
        p_gnu_io_elec["fomCosts"] = p_gnu_io_elec["fomCosts"]/100*p_gnu_io_elec["invCosts"]

        p_gnu_io_elec = p_gnu_io_elec.reset_index(names="unit")[["grid", "node", "unit", "input output",
                                                                 "conversionCoeff", "capacity", "unitSize", "vomCosts", "invCosts", "fomCosts", "annuityFactor", "techs"]]

        # part of p_gnu_io where grid = fuel
        p_gnu_io_fuel = p_gnu_io_elec[p_gnu_io_elec["conversionCoeff"] == 1 &
                                      ~p_gnu_io_elec["unit"].str.contains("charge")].copy()

        # set grid to fuel, set input output to input
        p_gnu_io_fuel["grid"] = "fuel"
        p_gnu_io_fuel["input output"] = "input"

        # change node to fuel node
        p_gnu_io_fuel["node"] = p_gnu_io_fuel["techs"]
        p_gnu_io_fuel.loc[p_gnu_io_fuel["node"].str.contains("CGT"), "node"] = "gas"

        # drop all costs columns, as the costs should only be declared once for output units
        p_gnu_io_fuel = p_gnu_io_fuel[["grid", "node", "unit", "input output", "conversionCoeff"]]

        # part of p_gnu_io where grid = storage
        p_gnu_io_storage = p_gnu_io_elec[p_gnu_io_elec["unit"].str.contains("charge")].copy()

        # set grid to storage, choose right input and output for charge and discharge units
        p_gnu_io_storage["grid"] = "storage"
        p_gnu_io_storage["input output"] = "input"
        p_gnu_io_storage.loc[p_gnu_io_storage["unit"].str.contains(" charge"), "input output"] = "output"

        # change node names
        p_gnu_io_storage["node"] = [name.rsplit(" ", 2)[0] for name in p_gnu_io_storage["unit"]]

        # add storage capacities
        if "storage_units_p_nom" in self.pypsaDs and not self.config["completeGreenField"]:
            capacity_storage_df = self.pypsaDs["storage_units_p_nom"].to_dataframe()\
                .rename(columns={"storage_units_p_nom": "capacity"})
            capacity_storage_df.index = capacity_storage_df.index + " charge unit"
            # drop all units which are not in p_gnu_io_storage
            capacity_storage_df = capacity_storage_df.loc[capacity_storage_df.index.isin(p_gnu_io_storage["unit"])]
            p_gnu_io_storage.set_index("unit", inplace = True)
            p_gnu_io_storage["capacity"] = capacity_storage_df["capacity"].combine_first(p_gnu_io_storage["capacity"])
            # drop all costs columns, as the costs should only be declared once for output units
            p_gnu_io_storage = p_gnu_io_storage.reset_index(names="unit")[["grid", "node", "unit", "input output", "conversionCoeff", 
                                                                           "capacity", "unitSize"]]
            p_gnu_io_storage.loc[p_gnu_io_storage["unit"].str.contains("discharge"), "capacity"] = 0
        else:
            # drop all costs columns, as the costs should only be declared once for output units
            p_gnu_io_storage = p_gnu_io_storage[["grid", "node", "unit", "input output", "conversionCoeff",
                                                "unitSize"]]
        
        # set upperLimitCapacityRatio for charge units in the storage grid
        p_gnu_io_storage.index = p_gnu_io_storage["node"]
        p_gnu_io_upperLimit = pd.concat([p_gnu_io_storage.loc[p_gnu_io_storage["unit"].str.contains(" charge")],
                                        self.pypsaDs["storage_units_max_hours"].to_dataframe()], axis=1).set_index("unit")
        if self.config["addAdvancedStorages"]: # only do it for hydro/PHS if using advanced storages
            p_gnu_io_upperLimit = p_gnu_io_upperLimit.loc[p_gnu_io_upperLimit.index.str.contains("hydro") 
                                                          | p_gnu_io_upperLimit.index.str.contains("PHS") ]
        p_gnu_io_storage.set_index("unit", inplace=True)
        p_gnu_io_storage = p_gnu_io_storage.merge(p_gnu_io_upperLimit, on=["grid", "node", "unit", "input output", 
                                                                        "conversionCoeff", "capacity", "unitSize"],
                                                how="left").rename({"storage_units_max_hours": "upperLimitCapacityRatio"}, axis=1)
        p_gnu_io_storage.reset_index(inplace=True)
        
        if self.config["addAdvancedStorages"]:
            storageUnits = [unit for unit in self.units if "storage" in unit]
            storageTechs = [tech for tech in self.techs if "storage" in tech]
            p_gnu_io_storage_add = self.costs.loc[storageTechs]
            p_gnu_io_storage_add = p_gnu_io_storage_add.reset_index(names="techs")
            p_gnu_io_storage_add["unit"] = storageUnits

            p_gnu_io_storage_add = p_gnu_io_storage_add.rename(columns={"FOM": "fomCosts", "VOM": "vomCosts",
                                                        "investment": "invCosts"})
    
            p_gnu_io_storage_add["annuityFactor"] = p_gnu_io_storage_add["discount rate"] / (1 - (1 + p_gnu_io_storage_add["discount rate"]) **
                                                                            (-p_gnu_io_storage_add["lifetime"]))
            p_gnu_io_storage_add.drop(columns=["CO2 intensity", "fuel", "efficiency",
                            "discount rate", "lifetime"], inplace=True)

            p_gnu_io_storage_add["grid"] = "storage"
            p_gnu_io_storage_add["node"] = [name.rsplit(" ", 2)[0] for name in p_gnu_io_storage_add["unit"]]
            p_gnu_io_storage_add["conversionCoeff"] = 1
            p_gnu_io_storage_add["unitSize"] = 1
            p_gnu_io_storage_add["capacity"] = "eps"
            p_gnu_io_storage_add["input output"] = "output"
            p_gnu_io_storage_add["upperLimitCapacityRatio"] = 1
            # FOM costs are given in % but must be absolute values for backbone
            p_gnu_io_storage_add["fomCosts"] = p_gnu_io_storage_add["fomCosts"]/100*p_gnu_io_storage_add["invCosts"]
            p_gnu_io_storage_add = p_gnu_io_storage_add[["grid", "node", "unit", "input output", "conversionCoeff",
                                                "capacity", "unitSize", "vomCosts", "invCosts", "fomCosts", "annuityFactor", 
                                                "upperLimitCapacityRatio"]] \
                                                    .set_index(["grid", "node", "unit", "input output"])
            p_gnu_io_storage.set_index(["grid", "node", "unit", "input output"], inplace=True)
            p_gnu_io_storage = pd.concat([p_gnu_io_storage, p_gnu_io_storage_add])
            p_gnu_io_storage.reset_index(inplace=True)

        # update of costs according to procedure in pypsa-eur

        # calculate connection costs between nodes
        def _prepare_connection_costs_per_link(n, costs, renewable_carriers, length_factor):
            if n.links.empty:
                return {}
            return {
                tech: (
                    n.links.length
                    * length_factor
                    * (
                        n.links.underwater_fraction * costs.loc[f"{tech}-connection-submarine", "discount rate"] / (1 - (1 + costs.loc[f"{tech}-connection-submarine", "discount rate"])
                                                                                                                    ** (-costs.loc[f"{tech}-connection-submarine", "lifetime"]))
                        * costs.at[tech + "-connection-submarine", "investment"]
                        + (1.0 - n.links.underwater_fraction) * costs.loc[f"{tech}-connection-underground", "discount rate"] / (1 - (1 + costs.loc[f"{tech}-connection-underground", "discount rate"])
                                                                                                                                ** (-costs.loc[f"{tech}-connection-underground", "lifetime"]))
                        * costs.at[tech + "-connection-underground", "investment"]
                    )
                )
                for tech in renewable_carriers
                if tech.startswith("offwind")
            }

        # update offshore wind costs with the connection costs
        def updateOffwindCosts(acdc, pypsa_path, network_unsimply, network_buses_sim, trafomap):
            # load offwind profiles from pypsa-eur
            profile_offwind_path = os.path.join(pypsa_path, f"resources/profile_offwind-{acdc}.nc")
            profile_offwind = xr.load_dataset(profile_offwind_path)

            # get average shore distance and underwater fraction from offwind profiles
            avDis = profile_offwind["average_distance"].to_dataframe()
            avDis.columns = ["investment"]
            underwaterFraction = profile_offwind["underwater_fraction"].to_dataframe()
            underwaterFraction.columns = ["investment"]

            # get weights (= size) from unsimpified network map
            weights = network_unsimply.generators["weight"]

            # calculate specific offshore costs for each node (with connection costs)
            # calculate differen annuity factors for turbine, station, underground and submarine connections
            afTur = self.costs.loc["offwind", "discount rate"] / (1 - (1 + self.costs.loc["offwind", "discount rate"])
                                                                  ** (-self.costs.loc["offwind", "lifetime"]))
            afStat = self.costs.loc[f"offwind-{acdc}-station", "discount rate"] / (1 - (1 + self.costs.loc[f"offwind-{acdc}-station", "discount rate"])
                                                                                   ** (-self.costs.loc[f"offwind-{acdc}-station", "lifetime"]))
            afConUn = self.costs.loc[f"offwind-{acdc}-connection-underground", "discount rate"] \
                / (1 - (1 + self.costs.loc[f"offwind-{acdc}-connection-underground", "discount rate"])
                   ** (-self.costs.loc[f"offwind-{acdc}-connection-underground", "lifetime"]))
            afConSub = self.costs.loc[f"offwind-{acdc}-connection-submarine", "discount rate"] \
                / (1 - (1 + self.costs.loc[f"offwind-{acdc}-connection-submarine", "discount rate"])
                   ** (-self.costs.loc[f"offwind-{acdc}-connection-submarine", "lifetime"]))
            # calculate total investment costs
            investOffshore = (self.costs.loc["offwind", "investment"] + 1/afTur*(afStat*self.costs.loc[f"offwind-{acdc}-station", "investment"]
                                                                                 + afConSub *
                                                                                 self.config["lines"]["length_factor"]*avDis
                                                                                 * underwaterFraction*self.costs.loc[f"offwind-{acdc}-connection-submarine", "investment"]
                                                                                 + afConUn*self.config["lines"]["length_factor"]*avDis*(1-underwaterFraction) *
                                                                                 self.costs.loc[f"offwind-{acdc}-connection-underground", "investment"]))
            investOffshore.index = pd.to_numeric(investOffshore.index)

            # add simplified buses to cost df
            investOffshore = pd.concat([network_buses_sim, investOffshore], axis=1).dropna()

            # calculate connection costs per link
            connection_costs_per_link = _prepare_connection_costs_per_link(
                network_unsimply, self.costs, [f"offwind-{acdc}"], self.config["lines"]["length_factor"]
            )

            # create graph from pyspa-eur network
            adj = network_unsimply.adjacency_matrix(
                weights=pd.concat(
                    dict(
                        Link=connection_costs_per_link[f"offwind-{acdc}"].reindex(network_unsimply.links.index),
                        Line=pd.Series(0.0, network_unsimply.lines.index),
                    )
                )
            )
            trafomap.name = "trafo_bus"
            trafomap.index = pd.to_numeric(trafomap.index)
            trafomap = pd.to_numeric(trafomap)

            combinedMap = pd.concat([trafomap, network_buses_sim], axis=1)
            combinedMap = combinedMap.loc[combinedMap["trafo_bus"]
                                          != combinedMap["bus_s"]].drop(columns=["trafo_bus"])
            # find cheapest way to connect the buses which were connected while simplifying
            for bus1, bus2 in zip(combinedMap.index, combinedMap["bus_s"]):
                if bus1 not in investOffshore.index:
                    continue
                dist_matrix = dijkstra(csgraph=adj, directed=False,
                                       indices=network_unsimply.buses.index.get_indexer([str(bus1)]))
                addCosts = dist_matrix[0, network_unsimply.buses.index.get_indexer([str(bus2)])][0]
                # add connection costs to new bus
                investOffshore.loc[bus1, "investment"] += addCosts/afTur

            # read in clustered network to find, which bus was connected to which node when clustering
            network_buses_clu_path = os.path.join(pypsa_path,
                                                  f"resources/busmap_elec_s_{self.numberOfBuses}.csv"
                                                  )
            network_buses_clu = pd.read_csv(
                network_buses_clu_path,
                names=["bus_s", "node"],
                header=1
            )
            # add bus weights to cost df
            weights = weights.to_frame().loc[weights.index.str.contains(f"offwind-{acdc}")]
            weights.index = [int(element.split(" ")[0]) for element in weights.index]
            investOffshore = pd.merge(investOffshore, weights.reset_index(names=["bus"]), on=["bus"])

            # add nodes to cost df
            investOffshore = pd.merge(network_buses_clu, investOffshore, on=["bus_s"])

            # calculate weighted average
            investOffshore["investment"] = investOffshore["investment"]*investOffshore["weight"]
            investOffshore = investOffshore.groupby("node").sum()
            investOffshore["investment"] = investOffshore["investment"]/investOffshore["weight"]

            return investOffshore.rename(columns={"investment": "invCosts"}).drop(columns=["bus_s", "weight"])

        # load unsimplified network map for offwind calculations
        network_unsimply_path = os.path.join(self.pypsa_path, "resources/networks/elec.nc")
        network_unsimply = pypsa.Network(network_unsimply_path)
        network_unsimply, trafomap = simplify_network_to_380(network_unsimply)

        # read in simplified network to find, which bus was connected to which new bus when simplifiying
        network_buses_sim_path = os.path.join(self.pypsa_path, "resources/busmap_elec_s.csv")
        network_buses_sim = pd.read_csv(
            network_buses_sim_path,
            names=["bus", "bus_s"],
            header=0,
            index_col=0
        )

        for acdc in ["ac", "dc"]:
            update = updateOffwindCosts(acdc, self.pypsa_path, network_unsimply, network_buses_sim, trafomap)
            p_gnu_io_elec = p_gnu_io_elec.set_index("node")
            p_gnu_io_elec.loc[p_gnu_io_elec["unit"].str.contains(f"offwind-{acdc}"), "invCosts"] \
                = update
            p_gnu_io_elec.reset_index(inplace=True)

        # update solar costs (pypsa takes half solar rooftop, half utility)
        def updateSolarCosts():
            update = pd.DataFrame()
            annuity_ro = self.costs.loc["solar-rooftop", "discount rate"] / \
                (1 - (1 + self.costs.loc["solar-rooftop", "discount rate"])
                 ** (-self.costs.loc["solar-rooftop", "lifetime"]))
            annuity_ut = self.costs.loc["solar-utility", "discount rate"] / \
                (1 - (1 + self.costs.loc["solar-utility", "discount rate"])
                 ** (-self.costs.loc["solar-utility", "lifetime"]))

            update.loc["solar", "investment"] = (self.costs.loc["solar-rooftop", "investment"]*annuity_ro
                                                 / annuity_ut+self.costs.loc["solar-utility", "investment"])/2
            update.loc["solar", "fomCosts"] = (self.costs.loc["solar-rooftop", "investment"]
                                               * self.costs.loc["solar-rooftop", "FOM"]/100
                                               + self.costs.loc["solar-utility", "investment"]
                                               * self.costs.loc["solar-utility", "FOM"]/100)/2
            update.name = "solar"

            return update.rename(columns={"investment": "invCosts"})[["invCosts", "fomCosts"]]

        update = updateSolarCosts()
        p_gnu_io_elec.loc[p_gnu_io_elec["techs"] == "solar", "invCosts"] = update.loc["solar", "invCosts"]
        p_gnu_io_elec.loc[p_gnu_io_elec["techs"] == "solar", "fomCosts"] = update.loc["solar", "fomCosts"]

        def updateStorageCosts(storageName, technology, p_gnu_io_elec):
            addCosts = self.costs.loc[storageName, "investment"]*self.pypsaDs["storage_units_max_hours"].to_dataframe()\
                .loc[p_gnu_io_elec.loc[p_gnu_io_elec["unit"].str.contains(f"{technology} discharge"), "node"] +
                     f" {technology}", "storage_units_max_hours"]
            addCosts.name = "invCosts"
            addCosts.index = addCosts.index + " discharge unit"
            annuityFactor = self.costs.loc[storageName, "discount rate"] / \
                (1 - (1 + self.costs.loc[storageName, "discount rate"])
                 ** (-self.costs.loc[storageName, "lifetime"]))
            p_gnu_io_elec = p_gnu_io_elec.set_index("unit")
            p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains(f"{technology} discharge"), "invCosts"] += \
                annuityFactor / \
                p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains(
                    f"{technology} discharge"), "annuityFactor"]*addCosts
            return p_gnu_io_elec.reset_index()

        if not self.config["addAdvancedStorages"]:
            for technology, storageName in zip(["battery", "H2"], ["battery storage", "Hydrogen-store"]):
                p_gnu_io_elec = updateStorageCosts(storageName, technology, p_gnu_io_elec)

            # shift electrolysis cost to storage grid
            p_gnu_io_elec.set_index("unit", inplace=True)
            p_gnu_io_storage.set_index("unit", inplace=True)
            p_gnu_io_storage.loc[p_gnu_io_storage.index.str.contains("H2 charge"), "invCosts"] = \
                p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "invCosts"]
            p_gnu_io_storage.loc[p_gnu_io_storage.index.str.contains("H2 charge"), "fomCosts"] = \
                p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "fomCosts"]
            p_gnu_io_storage.loc[p_gnu_io_storage.index.str.contains("H2 charge"), "annuityFactor"] = \
                p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "annuityFactor"]
            p_gnu_io_storage.loc[p_gnu_io_storage.index.str.contains("H2 charge"), "vomCosts"] = \
                p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "vomCosts"]
            p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "invCosts"] = ""
            p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "fomCosts"] = ""
            p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "annuityFactor"] = ""
            p_gnu_io_elec.loc[p_gnu_io_elec.index.str.contains("H2 charge"), "vomCosts"] = ""
            p_gnu_io_elec.reset_index(inplace=True)
            p_gnu_io_storage.reset_index(inplace=True)

        return pd.concat([p_gnu_io_elec, p_gnu_io_fuel, p_gnu_io_storage])[["grid", "node", "unit", "input output",
                                                                            "conversionCoeff", "capacity", "unitSize", "vomCosts", "invCosts", "fomCosts", "annuityFactor",
                                                                            "upperLimitCapacityRatio"]]

    def get_p_gnn(self):
        lengthFactor = self.config["lines"]["length_factor"]
        # read in all information on transfer lines
        lines_df = pd.concat([self.pypsaDs["lines_bus0"].to_dataframe().rename(columns={"lines_bus0": "from node"}),
                              self.pypsaDs["lines_bus1"].to_dataframe().rename(columns={"lines_bus1": "to node"}),
                              self.pypsaDs["lines_s_nom"].to_dataframe().rename(
                                  columns={"lines_s_nom": "transferCap"}),
                              self.pypsaDs["lines_length"].to_dataframe()], axis=1)

        # calculate investment costs based on line length
        lines_df["invCost_lines"] = lines_df["lines_length"] * \
            self.costs.loc["HVAC overhead", "investment"]*lengthFactor
        # as there is no possibility to add FOM, FOM is added to annuity factor
        lines_df["annuityFactor_lines"] = self.costs.loc["HVAC overhead", "FOM"]/100+self.costs.loc["HVAC overhead", "discount rate"] / \
            (1 - (1 + self.costs.loc["HVAC overhead", "discount rate"])
             ** (-self.costs.loc["HVAC overhead", "lifetime"]))
        lines_df.drop(columns=["lines_length"], inplace=True)

        # if backbone invest in lines is active, add flag expandLines to all countries which shall see expansion
        if self.config["investOptionsBackbone"]["lines"]["investOption"]:
            if self.config["investOptionsBackbone"]["lines"]["countries"] == "all":
                lines_df["expandLines"] = 1
            else:
                for country in self.config["investOptionsBackbone"]["lines"]["countries"]:
                    lines_df.loc[lines_df["from node"].str.contains(country), "expandLines"] = 1
                    lines_df.loc[lines_df["to node"].str.contains(country), "expandLines"] = 1
            lines_df.fillna(0, inplace=True)
        else:
            lines_df["expandLines"] = 0

        if "links_bus0" in self.pypsaDs:
            if "links_p_nom" in self.pypsaDs:
                links_df = pd.concat([self.pypsaDs["links_bus0"].to_dataframe().rename(columns={"links_bus0": "from node"}), self.pypsaDs["links_bus1"].to_dataframe()
                                      .rename(columns={"links_bus1": "to node"}), self.pypsaDs["links_p_nom"].to_dataframe()
                                      .rename(columns={"links_p_nom": "transferCap"}), self.pypsaDs["links_length"].to_dataframe(),
                                      self.pypsaDs["links_underwater_fraction"].to_dataframe()], axis=1)
            else:  # if there are only links with capacity 0
                links_df = pd.concat([self.pypsaDs["links_bus0"].to_dataframe().rename(columns={"links_bus0": "from node"}), self.pypsaDs["links_bus1"].to_dataframe()
                                      .rename(columns={"links_bus1": "to node"}), self.pypsaDs["links_length"].to_dataframe(),
                                      self.pypsaDs["links_underwater_fraction"].to_dataframe()], axis=1)
                links_df["transferCap"] = 0
            # link self.costs have three components: overhead and submarine parts and the inverter pair
            # all can have different annuityFactors / FOMs, must be added
            links_df["annuityFactor_links"] = self.costs.loc["HVDC overhead", "FOM"]/100 + self.costs.loc["HVDC overhead", "discount rate"] / \
                (1 - (1 + self.costs.loc["HVDC overhead", "discount rate"])
                 ** (-self.costs.loc["HVDC overhead", "lifetime"]))
            annuityFactor_submarine = self.costs.loc["HVDC submarine", "FOM"]/100 + self.costs.loc["HVDC submarine", "discount rate"] / \
                (1 - (1 + self.costs.loc["HVDC submarine", "discount rate"])
                 ** (-self.costs.loc["HVDC submarine", "lifetime"]))
            annuityFactor_inverter = self.costs.loc["HVDC inverter pair", "FOM"]/100 + self.costs.loc["HVDC inverter pair", "discount rate"] / \
                (1 - (1 + self.costs.loc["HVDC inverter pair", "discount rate"])
                 ** (-self.costs.loc["HVDC inverter pair", "lifetime"]))
            links_df["invCost_links"] = links_df["links_length"]*links_df["links_underwater_fraction"]*self.costs.at["HVDC submarine", "investment"]\
                * annuityFactor_submarine/links_df["annuityFactor_links"] + \
                links_df["links_length"]*(1-links_df["links_underwater_fraction"])*self.costs.at["HVDC overhead", "investment"] +\
                self.costs.at["HVDC inverter pair", "investment"] * \
                annuityFactor_inverter/links_df["annuityFactor_links"]
            links_df.drop(columns=["links_length", "links_underwater_fraction"], inplace=True)

            # if backbone invest option for links is active, add expandLinks flag to countries which shall see expansion
            if self.config["investOptionsBackbone"]["links"]["investOption"]:
                if self.config["investOptionsBackbone"]["links"]["countries"] == "all":
                    links_df["expandLinks"] = 1
                else:
                    for country in self.config["investOptionsBackbone"]["links"]["countries"]:
                        links_df.loc[links_df["from node"].str.contains(country), "expandLinks"] = 1
                        links_df.loc[links_df["to node"].str.contains(country), "expandLinks"] = 1
                links_df.fillna(0, inplace=True)
            else:
                links_df["expandLinks"] = 0
        else:
            links_df = pd.DataFrame()
            links_df["expandLinks"] = 0

        p_gnn_df = pd.concat([lines_df, links_df])

        # sort nodes such that node with smaller number/earlier in the alphabet comes earlier
        fromNode = [toNode if fromNode[0:3] > toNode[0:3] else fromNode for fromNode,
                    toNode in zip(p_gnn_df["from node"], p_gnn_df["to node"])]
        p_gnn_df["to node"] = [fromNode if fromNode[0:3] > toNode[0:3]
                               else toNode for fromNode, toNode in zip(p_gnn_df["from node"], p_gnn_df["to node"])]
        p_gnn_df["from node"] = fromNode
        fromNode = [toNode if fromNode[0:3] == toNode[0:3] and int(fromNode.split(" ")[1]) > int(toNode.split(" ")[1]) else fromNode for fromNode, toNode
                    in zip(p_gnn_df["from node"], p_gnn_df["to node"])]
        p_gnn_df["to node"] = [fromNode if fromNode[0:3] == toNode[0:3] and int(fromNode.split(" ")[1]) > int(toNode.split(" ")[1])
                               else toNode for fromNode, toNode in zip(p_gnn_df["from node"], p_gnn_df["to node"])]
        p_gnn_df["from node"] = fromNode

        # add lines/links between the same nodes
        p_gnn_df = p_gnn_df.groupby(["from node", "to node"]).agg({"transferCap": "sum", "invCost_lines": "mean",
                                                                   "annuityFactor_lines": "mean",
                                                                   "expandLines": "mean", "invCost_links": "mean",
                                                                   "annuityFactor_links": "mean", "expandLinks": "mean"})

        # take invCosts and annuityFactor from lines, if expansion is active else from links
        p_gnn_df["invCost"] = [p_gnn_df.loc[i, "invCost_lines"] if p_gnn_df.loc[i, "expandLines"] == 1 else
                               p_gnn_df.loc[i, "invCost_links"] if p_gnn_df.loc[i, "expandLinks"] == 1 else 0 for i in p_gnn_df.index]
        p_gnn_df["annuityFactor"] = [p_gnn_df.loc[i, "annuityFactor_lines"] if p_gnn_df.loc[i, "expandLines"] == 1 else
                                     p_gnn_df.loc[i, "annuityFactor_links"] if p_gnn_df.loc[i, "expandLinks"] == 1 else 0 for i in p_gnn_df.index]
        # add maximum capacity investment limit of infinity for all lines with expansion flag
        p_gnn_df["transferCapInvLimit"] = ["inf" if expandLines == 1 or expandLinks ==
                                           1 else 0 for expandLines, expandLinks in zip(p_gnn_df["expandLines"], p_gnn_df["expandLinks"])]

        p_gnn_df["availability"] = 1
        p_gnn_df["transferLoss"] = self.config["constants"]["transferLoss"]
        p_gnn_df["grid"] = "elec"

        p_gnn_df = p_gnn_df.reset_index()[["grid", "from node", "to node", "availability", "transferCap",
                                           "transferLoss", "invCost", "annuityFactor", "transferCapInvLimit"]]
        # add all lines with exchange from/to node to allow forward and backward transfer
        p_gnn_df = pd.concat([p_gnn_df, p_gnn_df.copy()
                              .rename(columns={"from node": "to node", "to node": "from node"})])

        return p_gnn_df

    def get_p_gn(self):
        p_gn_elec = pd.DataFrame()
        p_gn_elec["node"] = self.node
        p_gn_elec["grid"] = "elec"
        p_gn_elec["nodeBalance"] = 1

        p_gn_fuel = pd.DataFrame()
        p_gn_fuel["node"] = self.fuelNodes
        p_gn_fuel["grid"] = "fuel"
        p_gn_fuel["usePrice"] = 1

        p_gn_storage = pd.DataFrame()
        p_gn_storage["node"] = self.storage_nodes
        p_gn_storage["grid"] = "storage"
        p_gn_storage["nodeBalance"] = 1
        p_gn_storage["energyStoredPerUnitOfState"] = 1

        return pd.concat([p_gn_elec, p_gn_fuel, p_gn_storage]).fillna(0)[["grid", "node", "nodeBalance", "energyStoredPerUnitOfState", "usePrice"]]

    def get_p_gnBoundaryPropertiesForStates(self):
        # calculate maximum storage capacity by multiplying charge power with EP ratio
        p_gnBoundaryPropertiesForStates_up = self.pypsaDs["storage_units_max_hours"].to_dataframe()\
            .rename(columns={"storage_units_max_hours": "storage_units_p_nom"})\
            * self.pypsaDs["storage_units_p_nom"].to_dataframe()
        p_gnBoundaryPropertiesForStates_up = p_gnBoundaryPropertiesForStates_up.reset_index()\
            .rename(columns={"storage_units_i": "node", "storage_units_p_nom": "constant"})
        p_gnBoundaryPropertiesForStates_up = p_gnBoundaryPropertiesForStates_up\
            .loc[p_gnBoundaryPropertiesForStates_up["constant"] > 0]
        p_gnBoundaryPropertiesForStates_up["grid"] = "storage"
        p_gnBoundaryPropertiesForStates_up["useConstant"] = 1
        p_gnBoundaryPropertiesForStates_up["param_gnBoundaryTypes"] = "upwardLimit"

        p_gnBoundaryPropertiesForStates_down = pd.DataFrame()
        p_gnBoundaryPropertiesForStates_down["node"] = self.storage_nodes
        p_gnBoundaryPropertiesForStates_down["grid"] = "storage"
        p_gnBoundaryPropertiesForStates_down["useConstant"] = 1
        p_gnBoundaryPropertiesForStates_down["param_gnBoundaryTypes"] = "downwardLimit"

        p_gnBoundaryPropertiesForStates_spill = pd.DataFrame()
        p_gnBoundaryPropertiesForStates_spill["node"] = [node for node in self.storage_nodes if "hydro" in node]
        p_gnBoundaryPropertiesForStates_spill["grid"] = "storage"
        p_gnBoundaryPropertiesForStates_spill["useTimeseries"] = 1
        p_gnBoundaryPropertiesForStates_spill["param_gnBoundaryTypes"] = "maxSpill"

        return pd.concat([p_gnBoundaryPropertiesForStates_up, p_gnBoundaryPropertiesForStates_down,
                          p_gnBoundaryPropertiesForStates_spill]).fillna(0)[["grid", "node", "param_gnBoundaryTypes", "useConstant", "constant", "useTimeseries"]]

    def get_eff_lvl_group_unit(self):
        effLevelGroupUnit_df = pd.DataFrame(columns=["effLevel", "effSelector", "unit"])
        for level in ["level1", "level2", "level3"]:
            effLevel_df = pd.DataFrame(data=level, columns=["effLevel"], index=self.unit["unit"])
            # do not define effLevels for flowUnits
            effLevel_df = effLevel_df.loc[~effLevel_df.index.isin(
                self.pypsaDs["generators_t_p_max_pu_i"].to_dataframe().index)]
            effLevel_df["effSelector"] = "directOff"
            effLevelGroupUnit_df = pd.concat([effLevelGroupUnit_df, effLevel_df.reset_index(names="unit")])
        return effLevelGroupUnit_df

    def get_unitUnittype(self):
        return pd.DataFrame()
    # table made problems and is in my opinion very useless, so now it's empty
        # storageUnitUnittype = self.pypsaDs["storage_units_carrier"].to_dataframe()\
        #     .rename(columns={"storage_units_carrier": "generators_carrier"})
        # storageUnitUnittype.index = storageUnitUnittype.index + " discharge unit"
        # storageUnitUnittype2 = self.pypsaDs["storage_units_carrier"].to_dataframe()\
        #     .rename(columns={"storage_units_carrier": "generators_carrier"})
        # storageUnitUnittype2.index = storageUnitUnittype2.index + " charge unit"
        # unitUnittype = pd.concat([self.pypsaDs["generators_carrier"].to_dataframe(),
        #                           storageUnitUnittype, storageUnitUnittype2])\
        #     .reset_index(names="unit").rename(columns={"generators_carrier": "unittype"})
        # # add unittypes of additional invest units
        # investUnits = [unit for unit in self.units if "invest" in unit]
        # if len(investUnits) > 0:
        #     investUnittype = pd.DataFrame(investUnits, columns=["unit"])
        #     investUnittype["unittype"] = investUnittype["unit"].str.rsplit(" ", n=2, expand=True)[1]
        #     return pd.concat([unitUnittype, investUnittype])
        # else:
        #     return unitUnittype

    def get_grid(self):
        return pd.DataFrame(["elec", "fuel", "storage"], columns=["grid"])

    def get_gnss_bound(self):
        nodes_df = pd.DataFrame(data=self.storage_nodes, columns=["node"])
        grid_df = pd.DataFrame(data=["storage"]*len(self.storage_nodes), columns=["grid"])
        sample_df = pd.DataFrame(data=["s000"]*len(self.storage_nodes), columns=["sample"])

        return pd.concat([grid_df, nodes_df, sample_df, sample_df], axis=1, sort=False)

    def get_node(self):
        return pd.concat([self.node,  pd.DataFrame(self.fuelNodes, columns=["node"]), pd.DataFrame(self.storage_nodes, columns=["node"])])

    @property
    def unit(self):
        return pd.DataFrame(self.units, columns=["unit"])

    def get_unittype(self):
        return self.unittype

    @property
    def unittype(self):
        return self.pypsaDs["carriers_i"].to_dataframe().rename(columns={"carriers_i": "unittype"}).dropna().drop("")

    @property
    def node(self):
        return self.pypsaDs["buses_i"].to_dataframe().rename(columns={"buses_i": "node"})["node"]  # buses

    @property
    def fuelNodes(self):
        nonNodeUnittypes = ['onwind', 'offwind-dc', 'ror', 'OCGT', 'CCGT', 'solar', 'offwind-ac', 'hydro',
                            'PHS', 'H2', 'battery', 'AC', 'DC']
        fuelNodes = [unittype for unittype in self.unittype["unittype"] if not unittype in nonNodeUnittypes]
        if 'CCGT' in self.unittype["unittype"] or 'OCGT' in self.unittype["unittype"]:
            fuelNodes.append('gas')
        return fuelNodes

    @property
    def flows(self):
        return pd.DataFrame(["offwind-dc", "offwind-ac", "onwind", "solar", "ror"], columns=["flow"])

    # Write InvestInit fitting to pypsa-eur
    def writeInvestInit(self):
        conversion_path = Path(__file__)
        pp2bb_path = conversion_path.parent.parent
        with open(pp2bb_path.joinpath("resources/investInit_template.gms"), "r") as f:
            filedata = f.read()

        newdata = filedata.replace("pypsa-eur-length", str(self.datalength))
        newdata = newdata.replace("pypsa-eur-steplength", str(self.temporalResolution))

        with open("backbone_input/investInit.gms", "w") as f:
            f.write(newdata)


if __name__ == "__main__":

    pass
