'''
    pypsa2bb - create input data for the esm backbone, run backbone and analyse results
    Copyright (C) 2020-2022 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from copy import copy
from io import StringIO
import platform
from functools import partial, reduce
from itertools import groupby
import subprocess as sp
from pathlib import Path
import logging
from gams import SV_EPS, GamsDatabase, GamsSetRecord
from gams import GamsWorkspace, GamsException
import pandas as pd
import yaml
from collections import OrderedDict
import tsam.timeseriesaggregation as tsam


def configure_logger(name_of_script, snakemake=False, config=None):

    # get name for logger
    name = Path(name_of_script).name
    logger = logging.getLogger(name)

    if snakemake:
        logpath = snakemake.log[0]
        config = snakemake.config

    else:
        assert (
            config is not None
        ), "If called without snakemake, the config argument must be passed"
        name_of_log = Path(config["pypsa2bbPaths"]["pypsa_model_path"]).stem
        logpath = "logs/" + name_of_log + "/" + name + ".log"

    logfile = Path(logpath)

    # create directory if necessary
    log_dir = logfile.absolute().parent
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    # create new dict, to avoid storage of Filehandler objects
    default_config = {}
    default_config.update(config["logging"])
    default_config.update(
        {
            "handlers": [
                # Prefer the 'python' log, otherwise take the first log for each
                # Snakemake rule
                logging.FileHandler(logfile),
                logging.StreamHandler(),
            ]
        }
    )

    logging.basicConfig(**default_config)
    return logger


# taken from pypsa-eur
def mock_snakemake(file_py, **wildcards):
    """
    This function is expected to be executed from the 'scripts'-directory of '
    the snakemake project. It returns a snakemake.script.Snakemake object,
    based on the Snakefile.

    If a rule has wildcards, you have to specify them in **wildcards.

    Parameters
    ----------
    rulename: str
        name of the rule for which the snakemake object should be generated
    **wildcards:
        keyword arguments fixing the wildcards. Only necessary if wildcards are
        needed.
    """
    import snakemake as sm
    import os
    from pypsa.descriptors import Dict
    from snakemake.script import Snakemake

    rulename = Path(file_py).name.split(".")[0]
    script_dir = Path(file_py).parent.resolve()
    # root_dir = script_dir.parent.resolve()
    assert (
        Path.cwd().resolve() == script_dir
    ), f"mock_snakemake has to be run from the repository scripts directory {script_dir}"
    os.chdir(script_dir.parent)
    for p in sm.SNAKEFILE_CHOICES:
        if os.path.exists(p):
            snakefile = p
            break
    workflow = sm.Workflow(snakefile)
    workflow.include(snakefile)
    workflow.global_resources = {}
    rule = workflow.get_rule(rulename)
    dag = sm.dag.DAG(workflow, rules=[rule])
    wc = Dict(wildcards)
    job = sm.jobs.Job(rule, dag, wc)

    def make_accessable(*ios):
        for io in ios:
            for i in range(len(io)):
                io[i] = os.path.abspath(io[i])

    make_accessable(job.input, job.output, job.log)
    snakemake = Snakemake(
        job.input,
        job.output,
        job.params,
        job.wildcards,
        job.threads,
        job.resources,
        job.log,
        job.dag.workflow.config,
        job.rule.name,
        None,
    )
    # create log and output dir if not existent
    for path in list(snakemake.log) + list(snakemake.output):
        Path(path).parent.mkdir(parents=True, exist_ok=True)

    os.chdir(script_dir)
    return snakemake


def excel_to_frame_dict(xl_path):
    # read the excel file
    frame_dict = {}
    with pd.ExcelFile(xl_path) as xl:
        # iterate over sheets and write them as csv into csv_dir
        for name in xl.sheet_names:
            df = pd.read_excel(xl, name)
            frame_dict[name] = df

    return frame_dict


def excel_to_csvs(xl_path, csv_dir_parent="backbone_input/versioned", overwrite=False):
    """converts an excel file to a directory of .csv files where each .csv filename corresponds to a sheet of the excel file

    Args:
        xl_path (str): Path to the excel file
        csv_dir (str, optional): Path to the directory in which the csv directory will be created. Name of the csv_dir will be derived from exel file name.
        overwrite (bool, optional): Wether or not to overwrite existing files. Defaults to False.
    """
    # if no specific directory was passed, create directory name from filename
    csv_dir_name = Path(xl_path).stem
    csv_dir = Path(csv_dir_parent).joinpath(csv_dir_name).resolve()

    # create directory
    csv_dir.mkdir(exist_ok=overwrite, parents=True)

    # read excel contents
    frame_dict = excel_to_frame_dict(xl_path)

    # write contents to csvs
    frame_dict_to_csvs(frame_dict, csv_dir)


def excel_to_gdx(xl_path, gdx_path=None):
    """converts an excel file to a gdx file (both could be used as backbone inputs)

    Args:
        xl_path (str): path to excel file
        gdx_path (str): path to gdx file
    """
    if gdx_path is None:
        gdx_path = Path(xl_path).as_posix().replace(".xlsx",".gdx")

    if platform.system() == "Windows":
        sp_output = sp.run(["gdxxrw",xl_path,f"output={gdx_path}","index=index!"], stdout=sp.PIPE)
        assert sp_output.returncode == 0, \
             AssertionError(f"gdxxrw returned {sp_output.returncode}. Output was:\n{sp_output.stdout.decode()}")
    else:
        frames = excel_to_frame_dict(xl_path)
        frame_dict_to_gdx(frames, gdx_path)

    return gdx_path


def csvs_to_frame_dict(csv_dir):
    dir_path = Path(csv_dir)
    files = [f for f in dir_path.iterdir() if f.name.endswith(".csv")]

    frame_dict = {}
    for file in files:
        try:
            df = pd.read_csv(file)
            # if a column contains non-numeric values e.g. "eps", it is parsed as str
            # strings are written to excel with preceding ' which results in a backbone error
            for col in df.columns:
                # find indices of eps
                if "eps" in df[col].values:
                    soon_zeros = df[col] == "eps"
                    # set others to be floats
                    df[col] = df[col][~soon_zeros].astype(float)

        except pd.io.parsers.EmptyDataError:
            print(f"{file} is empty.")
            df = pd.DataFrame()
        except Exception as e:
            e.args = (f"error occured with {file}:", *e.args)
            raise e  # Exception(f"error occured with {file}:", e.with_traceback())
        frame_dict[file.stem] = df

    return frame_dict


def csvs_to_excel(csv_dir, xl_path=None, overwrite=False):
    dir_path = Path(csv_dir)

    frame_dict = csvs_to_frame_dict(csv_dir)

    if not xl_path:
        xl_path = Path("results").joinpath(f"{dir_path.name}.xlsx")

    frame_dict_to_excel(frame_dict, xl_path)


def frame_dict_to_excel(frame_dict, xl_path):
    with pd.ExcelWriter(xl_path) as writer:
        for key in frame_dict:
            frame_dict[key].to_excel(writer, sheet_name=key, index=False)


def frame_dict_to_csvs(
    frames_dict, csv_dir="backbone_input/versioned/default", overwrite=True
):

    csv_dir = Path(csv_dir).resolve()
    csv_dir.mkdir(exist_ok=overwrite, parents=True)

    for key in frames_dict.keys():
        frames_dict[key].to_csv(f"{csv_dir}/{key}.csv", index=False)
    print("written frames as csvs to ", csv_dir)


def frame_dict_to_gdx(frames_dict, gdx_path, ws_dir=Path(__file__).parent.parent):
    """Writes the dict of pandas.DataFrames to a gdx file. 

    Args:
        frames_dict (dict): dict of pd.DataFrames
        gdx_path (str): 

    """
    # GamsWorkspace needs to be initialized so that relative paths supplied by snakemake workout
    ws = GamsWorkspace(ws_dir)
    db = ws.add_database()
    lookup_df = frames_dict["index"].set_index("Symbol")

    sheet_diff = set(frames_dict.keys()).difference(frames_dict["index"]["Symbol"])
    assert sheet_diff == {"index"}, AssertionError(
        f"Expected to find only a missing index sheet, but got {sheet_diff}"
    )

    for sheet_key in lookup_df.index:
        gams_type = lookup_df.loc[sheet_key, "Type"]

        if gams_type.lower() == "par":
            # print(sheet_key)
            add_param_to_gdx_db(sheet_key, frames_dict, db)
        else:
            # print(sheet_key)
            add_set_to_gdx_db(sheet_key, frames_dict, db)

    db.export(gdx_path)


def add_param_to_gdx_db(sheet_name, frames_dict, gams_db, debug=False):
    """
        in backbone, the sheet_name of the excel file (= the file name of the csv) 
        corresponds to the parameter/set name.
    """
    lookup_df = frames_dict["index"].set_index("Symbol")

    r_dim = lookup_df.loc[sheet_name, ["Rdim", "Cdim"]].sum().astype(int)
    r_dim = int(r_dim)  # gams doesn't like numpy.int32 from above
    # add 1 to dimension for values
    if sheet_name in [
        "p_groupPolicyUnit",
        "p_s_discountFactor",
        "p_groupPolicyEmission",
    ]:
        # theres something funky going on with these two parameters
        slice_mod = 1
    else:
        slice_mod = 0
    param = gams_db.add_parameter(sheet_name, r_dim)
    for index, row in frames_dict[sheet_name].iterrows():
        # index should correspond to dimensionality
        # e.g. for p_gnn -> this is grid,node,node (3 dimensions)
        # any entries with index higher than dimension are corresponding values e.g. "transfercap"
        dim_index = tuple(row[: r_dim - 1 + slice_mod].values)
        for column_key, val in row[r_dim - 1 + slice_mod:].items():
            try:
                value_index = dim_index + (column_key,)
                if debug:
                    print(
                        sheet_name, dim_index, column_key, val, f"dim={r_dim}", sep=", "
                    )

                if slice_mod:
                    # switch up index for weird params
                    value_index = dim_index
                if val == "inf":
                    param.add_record(value_index).value = float("inf")
                elif val == "eps":
                    param.add_record(value_index).value = SV_EPS
                elif pd.isna(val):
                    # ignore nans
                    continue
                elif not pd.isna(val) and type(val) != str:
                    # only set value if it is neither nan nor str
                    param.add_record(value_index).value = val
                elif val != "":
                    # some unhandled value
                    raise ValueError(
                        f"Value of type {type(val)} with value {val} not supported"
                    )
                else:
                    # ignore empty strings
                    pass
            except Exception as e:
                print(sheet_name, dim_index, column_key, val, f"dim={r_dim}", sep=", ")
                raise e


def add_set_to_gdx_db(sheet_name, frames_dict, gams_db, debug=False):
    lookup_df = frames_dict["index"].set_index("Symbol")

    r_dim = lookup_df.loc[
        sheet_name, "Rdim"
    ]  # lookup_df.loc[sheet_name,["Rdim","Cdim"]].sum().astype(int)
    r_dim = int(r_dim)  # gams doesn't like numpy.int32 from above
    gams_set = gams_db.add_set(sheet_name, r_dim,)
    for index, row in frames_dict[sheet_name].iterrows():
        dim_index = tuple(row[:r_dim].values)
        if debug:
            print(sheet_name, dim_index)
        try:
            gams_set.add_record(dim_index)
        except Exception as e:
            print(sheet_name, dim_index)
            raise e


def ordered_load(stream, Loader=yaml.SafeLoader, object_pairs_hook=OrderedDict):
    # helper function to read yaml and circumvent issues of OrderedDict
    # Source: https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )
    return yaml.load(stream, OrderedLoader)


def ordered_dump(data, stream=None, Dumper=yaml.SafeDumper, **kwds):
    # helper function to write yaml and circumvent issues of OrderedDict
    # Source: https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
    class OrderedDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items()
        )

    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **kwds)


class BackboneInput:
    def __init__(self, path):
        # this section is redundant with results.BackboneResult
        ws = GamsWorkspace(".")

        self._path = Path(path).absolute()

        if not self._path.exists():
            raise(FileNotFoundError(self._path))

        self.gams_db = ws.add_database_from_gdx(self._path.as_posix())

        repo_root = Path(__file__).parents[1]

        bb_index = pd.read_csv(repo_root.joinpath("resources/index_file.csv"), sep=";")
        self.input_symbols = bb_index["Symbol"]
        self.set_symbols_as_attribs()

    def set_symbols_as_attribs(self):
        """Sets all input symbols as methods. Allows for easy access like `bb_input.p_gnu_io()` or `bb_input.ts_influx()` 
        """
        for symbol in self.input_symbols:

            setattr(self, symbol, partial(self.param_as_df_gdxdump, symbol))

    def param_as_df_gdxdump(self, symbol, encoding="1252", convert_time=True) -> pd.DataFrame:
        """ Use the 'gdxdump' GAMS utility via subprocess to convert a parameter into a pd.DataFrame.
        This is sometimes beneficial, to circumvent decoding errors
        """
        gdxdump = sp.run(
            ["gdxdump", self._path, "format", "csv", "symb", symbol], stdout=sp.PIPE
        )
        csv_data = gdxdump.stdout.decode(encoding=encoding)
        header = csv_data.partition("\n")[0]

        header = [x.strip('"') for x in header.split(",")]
        dtypes = [str if x != "Val" else float for x in header]
        dtypes = dict(zip(header, dtypes))
        df = pd.read_csv(StringIO(csv_data), dtype=dtypes, na_values="Eps")

        # converts the time column to int by omitting the first character
        if convert_time:
            if "t" in df.columns:
                df["t"] = df["t"].apply(lambda t: int(t[1:]))
            else:
                # for backbone inputs, the index is ["Dim1", "Dim2", ... , "Val"], but there might be a time column
                first_row = df.loc[0]
                is_time = first_row.apply(lambda x: str(x).startswith("t000"))

                if any(is_time):
                    # any is necessary, otherwise *.idxmax()  simply returns first element
                    time_col = is_time.idxmax()  # should be something like "Dim4"
                    df[time_col] = df[time_col].apply(lambda t: int(t[1:]))
                    df.rename({time_col: "t"}, axis=1, inplace=True)

        return df

    def update_gams_parameter_in_db(
        self,
        param_name: str,
        indices: list,
        value: float = None,
        apply_percentage: float = None,
    ) -> GamsDatabase:
        # note that you are always only working on the same instance of the database.
        # retrieve the parameter i.e. "p_groupPolicyEmission"
        parameter = self.gams_db.get_symbol(param_name)

        # retrieve the record from that parameter i.e. ["emission group", "emissionCap", "CO2"]
        try:
            record = parameter.find_record(indices)
        except GamsException as ge:
            if "Cannot find" in ge.value:
                # This is potentially dangerous since any index value can be added.
                # However, in the case of a bad index, GAMS should throw a DomainViolationError
                print(f"Warning: Caught 'GamsException: {ge.value}'.\n\tSetting {indices} to {value}")
                record = parameter.add_record(indices)
            else:
                raise ge
        # add_record
        if not apply_percentage and type(value) is not None:
            if isinstance(record, GamsSetRecord):
                # sets only have text i.e. "Y"
                record.set_text(value)
            else:
                record.set_value(value)
        elif apply_percentage:
            current_val = record.get_value()
            record.set_value(current_val * apply_percentage)
        else:
            raise ValueError("one of 'value' and 'apply_percentage' must be passed!")
        return self.gams_db


    def aggregate_timeseries(self, export_path, hours_per_period=24*7, n_periods=3):
        ts_params = ["ts_influx", "ts_node", "ts_unit", "ts_cf"]

        ts_frames = []
        # retrieve and transform timeseries data (not exaustive)
        for p in ts_params:
            p_df = self.param_as_df_gdxdump(p)

            # index of the time column
            time_col = list(p_df.columns).index("t")

            # all the previous columns (time is usually the penultimate column)
            index_cols = list(p_df.columns[:time_col])

            # transform from long to wide dataformat
            p_df = p_df.pivot(index="t", columns=index_cols, values="Val")

            # join column names from different levels and prepend param_name
            # these will come in handy, when writing aggregated data to the *.gdx
            # col is tuple, thus must be converted to a list to be mutable
            p_df.columns = ["|".join([p] + list(col)) for col in p_df.columns]

            ts_frames.append(p_df)
        # merge ts_frames to a single dataframe
        ts_df = reduce(lambda left, right: pd.merge(
            left, right, left_index=True, right_index=True, how="outer"), ts_frames)

        # create and set a datetimeindex
        dateindex = pd.date_range(start="2020-01-01 00:00", periods=8760, freq="H")
        ts_df.set_index(dateindex, inplace=True)
        ts_df.fillna(0, inplace=True)

        # create aggregation
        aggregation = tsam.TimeSeriesAggregation(
            ts_df,
            hoursPerPeriod=hours_per_period,
            noTypicalPeriods=n_periods,
            clusterMethod="hierarchical",
            representationMethod="distributionRepresentation",
            distributionPeriodWise=False,
            rescaleClusterPeriods=False,
        )

        # calculate the typical periods
        typeriods = aggregation.createTypicalPeriods()

        # get the sequence and number of occurances
        period_sequence = [(k, sum(1 for i in g)) for k, g in groupby(aggregation.clusterOrder)]

        # concatenate
        ts_input_data = pd.concat(
            [typeriods.loc[period_no, :] for period_no, _ in period_sequence]
        ).reset_index(drop=True)

        # write data to gdx
        for col_name in ts_input_data.columns:
            print("updating", col_name)

            # recreate the indices from the column names
            bb_indices = col_name.split("|")

            for i, val in enumerate(ts_input_data[col_name].values):
                timestep = f"t{i+1:06}"
                self.update_gams_parameter_in_db(bb_indices[0], bb_indices[1:]+[timestep], val)

        # add samples to .gdx where nessesary
        samples_and_lengths = [{"name": f"s{i:03}", "length": hours_per_period, "weight": tup[1], "cluster_no": tup[0]}
                               for i, tup in enumerate(period_sequence)]
        sample_df = pd.DataFrame.from_records(samples_and_lengths, index="name")
        sample_names = sample_df.index.to_list()

        for sample in sample_df.index:
            self.update_gams_parameter_in_db("p_s_discountFactor", [sample], value=1)
            # : ["sample",	"group"]
            self.update_gams_parameter_in_db("sGroup", [sample, "emission group"], value="Y")
            assert "emission group" in self.group().values

        if len(self.group().values) > 2:
            print(
                f"Warning: samples have been added for group 'emission group' but more are present in the .gdx:{self.group().values}")

        # bind samples in gnss_bound
        # get nodes that have a state
        p_gn = self.p_gn()
        nodes_w_state = p_gn.query("Dim3=='energyStoredPerUnitOfState'")[["Dim1", "Dim2"]]
        # nodes_w_state.append([nodes_w_state]*(len(sample_names)-1))
        gnss_df = pd.concat([nodes_w_state]*len(sample_names))
        gnss_df.sort_values(by="Dim2", inplace=True)

        from_sample = []
        to_sample = []
        for i in range(len(gnss_df)):
            from_sample.append(sample_names[i % len(sample_names)])
            to_sample.append(sample_names[(i+1) % len(sample_names)])

        gnss_df["from_sample"] = from_sample
        gnss_df["to_sample"] = to_sample

        # clear records in bounds, because they might be invalid post-aggregation
        gnss_bounds = self.gams_db.get_symbol("gnss_bound")
        gnss_bounds.clear()

        for gnss_indices in gnss_df.values:
            # print(gnss_indices)
            self.update_gams_parameter_in_db("gnss_bound", list(gnss_indices), value="Y")

        print("Dont forget to update your bb input configuration!!")
        for i, tup in enumerate(period_sequence):
            print(f"s{i:03} - cluster:{tup[0]} - weight:{tup[1]},\tlength: {hours_per_period}")

        self.gams_db.export(export_path)
        pass

def run_backbone(
    input_dir,
    input_file_gdx,
    output_dir,
    output_file,
    backbone_dir=None,
    **kwargs,
):
    """runs Backbone with the specified arguments

    Args:
        input_dir (str): Path to the input directory for Backbone to read configurations (e.g. `investInit.gms`)
        input_file_gdx (str): Path to the Backbone input file
        output_dir (str): Path to the output directory, where results shall be written
        output_file (str): Name of the Backbone result file
        backbone_dir (str, optional): Path to the Backbone Framework. Defaults to the path of the submodule in this repository.
    """

    # if not specified, set backbone_dir to submodule path
    if not backbone_dir:
        backbone_dir = Path(__file__).parent.parent.joinpath("backbone")

    # keyword arguments are parsed into backbone suitable form
    # e.g. dict(maxTotalCost=42) will be passed as "--maxTotalCost=42" to backbone
    keyword_args = [f"--{k}={v}" for k, v in kwargs.items()]
    
    # spaces in file names don't work through subprocesses
    contain_spaces = [" " in str(file_or_dir) for file_or_dir in [input_dir, input_file_gdx, output_dir, output_file]]
    if any(contain_spaces):
        raise ValueError("Passing paths with spaces via subprocess to Gams is not supported yet.")

    # absolute paths are better for gams
    input_dir = Path(input_dir).absolute().as_posix()
    input_file_gdx = Path(input_file_gdx).absolute().as_posix()
    output_dir = Path(output_dir).absolute().as_posix()

    process_output = sp.run(
        [
            "gams",
            "Backbone.gms",
            f"--input_dir={input_dir}",
            f"--input_file_gdx={input_file_gdx}",
            f"--output_dir={output_dir}",
            f"--output_file={output_file}",
            *keyword_args,
        ],
        cwd=backbone_dir,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        # text= True
    )

    # stdout, stderr = process_output.communicate()
    # write terminal output to a file
    open(f"{output_dir}/{Path(output_file).stem}_terminal.log", "w").write(
        process_output.stdout.decode()
    )
    open(f"{output_dir}/{Path(output_file).stem}_error.log", "w").write(
        process_output.stderr.decode()
    )
    print("finished", output_file)
    return Path(output_dir).joinpath(output_file).absolute()



if __name__ == "__main__":
    bb_input = BackboneInput("results/SFH_simple_batt_clo0.5 copy.gdx")
    # bb_input.aggregate_timeseries("test.gdx")
    bb_input.update_gams_parameter_in_db("sGroup", ["s001", "emission group"], value="Y")
    bb_input.gams_db.export("test.gdx")
    # input_file = BackboneInput(r"path/to/input/with/emission/cap.gdx")
    # p_gnu_io = input_file.p_gnu_io()
    # print(p_gnu_io)

    # # create a sensitivity analysis:
    # for percentage in [0.8, 0.6, 0.4, 0.2]:

    #     new_db = input_file.update_gams_parameter_in_db(
    #         "p_groupPolicyEmission",
    #         ["emission group", "emissionCap", "CO2"],
    #         apply_percentage=percentage,
    #     )

    #     new_db.export(f"modified_input_cap_{percentage}.gdx")
    # pass
