'''
    pypsa2bb - create input data for the esm backbone, run backbone and analyse results
    Copyright (C) 2020-2022 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import pandas as pd
from conversion import NC2XLS_Converter
from helpers import configure_logger, mock_snakemake
import sys
import numpy as np
import os
import geopandas as gpd


# check if snakemake variable is present
if "snakemake" in locals():
    # this script has been called as part of a snakemake workflow
    # modify input and output paths accordingly
    logger = configure_logger(snakemake.rule, snakemake=snakemake)
    config = snakemake.config
    config["pypsa2bbPaths"]["output_path"] = snakemake.output[0]
    config["pypsa2bbPaths"]["pypsa_model_path"] = snakemake.input[0]
elif len(sys.argv) > 1:
    # this script has been called from the command line
    # todo: implement a way to call the script without snakemake maybe by mocking it
    from constants import config

    config["pypsa2bbPaths"]["pypsa_model_path"] = sys.argv[1]
    config["pypsa2bbPaths"]["output_path"] = "/".join(sys.argv[2:])
    logger = configure_logger(__file__, config=config)
else:
    # this script has been called without arguments
    from constants import config

    logger = configure_logger(__file__, config=config)
    logger.warning(
        "pypsa_to_backbone.py called without snakemake or args. All paths will be derived from config."
    )

# store current config for debugging
# json.dump(config, open(config_path, "w"))

# create instance of converter, this contains config
nc_converter = NC2XLS_Converter(config)


# create the frames dict, containing a dataframe for each row in the index file
index_df = pd.read_csv(config["pypsa2bbPaths"]["index_path"], sep=";")
frames = {}
frames["index"] = index_df
for key in index_df["Symbol"]:
    frames[key] = pd.DataFrame(data={}, columns=[key])

# grid
# TODO: we should define a translation table: pypsa_names -> bb_names
frames["grid"] = nc_converter.get_grid()

# nodes = buses + fuels
frames["node"] = nc_converter.get_node()

# flow
frames["flow"] = nc_converter.flows

# unittype = carriers
frames["unittype"] = nc_converter.get_unittype()

# unit = generators
frames["unit"] = nc_converter.unit

# unitUnittype
frames["unitUnittype"] = nc_converter.get_unitUnittype()

# effLevelGroupUnit
frames["effLevelGroupUnit"] = nc_converter.get_eff_lvl_group_unit()

# group
frames["group"] = nc_converter.get_group()

# gngroup
frames["gngroup"] = nc_converter.get_gngroup()

# gnugroup
frames["gnugroup"] = nc_converter.get_gnugroup()

# sgroup
frames["sgroup"] = nc_converter.get_sgroup()

# ugroup
frames["ugroup"] = nc_converter.get_ugroup()

# p_gn
frames["p_gn"] = nc_converter.get_p_gn()

# p_gnBoundaryPropertiesForStates
frames[
    "p_gnBoundaryPropertiesForStates"
] = nc_converter.get_p_gnBoundaryPropertiesForStates()

# p_gnn
frames["p_gnn"] = nc_converter.get_p_gnn()

# ts_gnn
frames["ts_gnn"] = pd.DataFrame(data=[], columns=["grid", "node", "node", "param_gnn", "forecast index","t00000"])

# p_gnu_io
frames["p_gnu_io"] = nc_converter.get_p_gnu_io()

# p_unit
frames["p_unit"] = nc_converter.get_p_unit()

# flowUnit
frames["flowUnit"] = nc_converter.get_flowunit()

# emission
frames["emission"]["emission"] = ["CO2"]

# p_nemission
frames["p_nEmission"] = nc_converter.get_p_nemission()

# ts_PriceChange
frames["ts_PriceChange"] = nc_converter.get_ts_PriceChange()

# ts_emissionPriceChange
frames["ts_emissionPriceChange"] = nc_converter.get_ts_emissionPriceChange()

# ts_cf
frames["ts_cf"] = nc_converter.get_ts_cf()

# ts_influx and ts_node
frames["ts_influx"], frames["ts_node"] = nc_converter.get_ts_influx_and_ts_node()

# t_invest
frames["t_invest"]["t_invest"] = ["t000001"]

# utAvailabilityLimit
frames["utAvailabilityLimits"] = nc_converter.get_availability_limits()

# p_groupPolicy
frames["p_groupPolicy"] = nc_converter.get_p_groupPolicy()

# p_groupPolicyUnit
frames["p_groupPolicyUnit"] = nc_converter.get_p_groupPolicyUnit()

# p_groupPolicyEmission
frames["p_groupPolicyEmission"] = nc_converter.get_p_groupPolicyEmission()

# ts_unit
frames["ts_unit"] = pd.DataFrame(data=[], columns=["unit", "param_unit", "forecast index","t00000"])

# p_unitConstraint
frames["p_unitConstraint"] = pd.DataFrame(data=[], columns=["unit", "constraint", "values"])

# p_unitConstraintNode
frames["p_unitConstraintNode"] = pd.DataFrame(data=[], columns=["unit", "constraint", "node", "values"])

# restype
frames["restype"] = pd.DataFrame(
    data=["primary", "secondary", "tertiary"], columns=["restype"]
)
# gn2ngroup
frames["gn2ngroup"] = pd.DataFrame(data=[], columns=["grid", "node", "node", "group"])

# gnss_bound
frames["gnss_bound"] = nc_converter.get_gnss_bound()

nc_converter.writeInvestInit()  # write fitting investInit in backbone Folder

# combine offwind ac and dc
if config["combineOffwindAcAndDc"]:
    unitDict_df = frames["unit"]["unit"].str.replace("-ac", "").str.replace("-dc", "").to_frame(name = "newUnit")
    unitDict_df.index = frames["unit"]["unit"]

    flowDict_df = frames["flow"]["flow"].str.replace("-ac", "").str.replace("-dc", "").to_frame(name = "newFlow")
    flowDict_df.index = frames["flow"]["flow"]

    frames["unit"].index = frames["unit"]["unit"]
    frames["unit"] = unitDict_df["newUnit"].combine_first(frames["unit"]["unit"]).drop_duplicates().to_frame(name="unit")
    
    frames["effLevelGroupUnit"]["unit"] = frames["effLevelGroupUnit"]["unit"].replace(unitDict_df["newUnit"])
    frames["effLevelGroupUnit"] = frames["effLevelGroupUnit"].drop_duplicates(["effLevel", "unit"])

    frames["p_gnu_io"]["unit"] = frames["p_gnu_io"]["unit"].replace(unitDict_df["newUnit"])
    frames["p_gnu_io"]["capacity"] = frames["p_gnu_io"]["capacity"].replace("eps", 0)
    frames["p_gnu_io"] = frames["p_gnu_io"].groupby(["grid", "node", "unit", "input output"]).agg({"conversionCoeff": "mean", "capacity": "sum", "unitSize": "mean", "vomCosts": "mean", "invCosts": "mean", "fomCosts": "mean", "annuityFactor": "mean",
                                                                                "upperLimitCapacityRatio": "mean"})
    frames["p_gnu_io"] = frames["p_gnu_io"].reset_index()
    frames["p_gnu_io"].loc[frames["p_gnu_io"]["input output"] == "output", "capacity"] = frames["p_gnu_io"].loc[frames["p_gnu_io"]["input output"] == "output", "capacity"].replace(0, "eps")

    frames["p_unit"]["unit"] = frames["p_unit"]["unit"].replace(unitDict_df["newUnit"])
    frames["p_unit"]["eff00"] = frames["p_unit"]["eff00"].replace("", 0)
    frames["p_unit"]["maxUnitCount"] = frames["p_unit"]["maxUnitCount"].replace("inf", np.inf)
    frames["p_unit"] = frames["p_unit"].groupby("unit").agg({"eff00": "mean", "availability": "first", "maxUnitCount": "sum"}).reset_index()
    frames["p_unit"]["maxUnitCount"] = frames["p_unit"]["maxUnitCount"].replace(np.inf, "inf")

    frames["flow"].index = frames["flow"]["flow"]
    frames["flow"] = flowDict_df["newFlow"].combine_first(frames["flow"]["flow"]).drop_duplicates().to_frame(name="flow")

    frames["flowUnit"]["flow"] = frames["flowUnit"]["flow"].replace(flowDict_df["newFlow"])
    frames["flowUnit"]["unit"] = frames["flowUnit"]["unit"].replace(unitDict_df["newUnit"])
    frames["flowUnit"] = frames["flowUnit"].drop_duplicates(["flow", "unit"])

    frames["ts_cf"]["flow"] = frames["ts_cf"]["flow"].replace(flowDict_df["newFlow"])
    frames["ts_cf"] = frames["ts_cf"].groupby(["flow", "node", "forecast index"]).mean().reset_index()

    frames["utAvailabilityLimits"]["unit"] = frames["utAvailabilityLimits"]["unit"].replace(unitDict_df["newUnit"])
    frames["utAvailabilityLimits"] = frames["utAvailabilityLimits"].drop_duplicates(["unit"])

    frames["ugroup"]["unit"] = frames["ugroup"]["unit"].replace(unitDict_df["newUnit"])
    frames["ugroup"] = frames["ugroup"].drop_duplicates(["unit", "group"])

    frames["gnugroup"]["unit"] = frames["gnugroup"]["unit"].replace(unitDict_df["newUnit"])
    frames["gnugroup"] = frames["gnugroup"].drop_duplicates(["node", "unit", "group"])

    frames["p_groupPolicyUnit"]["unit"] = frames["p_groupPolicyUnit"]["unit"].replace(unitDict_df["newUnit"])
    frames["p_groupPolicyUnit"] = frames["p_groupPolicyUnit"].drop_duplicates(["group", "unit"])

# perform aggregation afterwards
if config["aggregateNodes"]:
    nodesToAggregate = config["nodesToAggregate"]
    nodeDict_df = pd.DataFrame()
    for newNode in nodesToAggregate.keys():
        for oldNode in nodesToAggregate[newNode]:
            oldNodes = frames["node"].loc[frames["node"]["node"].str.contains(oldNode)]
            oldNodes = pd.concat([oldNodes, oldNodes.rename(columns = {"node": "newNode"})["newNode"].str.replace(oldNode, newNode)], axis = 1)
            nodeDict_df = pd.concat([nodeDict_df, oldNodes])
    nodeDict_df.set_index("node", inplace = True)

    unitDict_df = pd.DataFrame()
    for newNode in nodesToAggregate.keys():
        for oldNode in nodesToAggregate[newNode]:
            oldUnits = frames["unit"].loc[frames["unit"]["unit"].str.contains(oldNode)] 
            oldUnits = pd.concat([oldUnits, oldUnits.rename(columns = {"unit": "newUnit"})["newUnit"].str.replace(oldNode, newNode)], axis = 1)
            unitDict_df = pd.concat([unitDict_df, oldUnits])
    unitDict_df.set_index("unit", inplace = True)

    groupDict_df = pd.DataFrame()
    for newNode in nodesToAggregate.keys():
        for oldNode in nodesToAggregate[newNode]:
            oldGroups = frames["group"].loc[frames["group"]["group"].str.contains(oldNode)] 
            oldGroups = pd.concat([oldGroups, oldGroups.rename(columns = {"group": "newGroup"})["newGroup"].str.replace(oldNode, newNode)], axis = 1)
            groupDict_df = pd.concat([groupDict_df, oldGroups])

            #groups with only country codes
            oldGroups = frames["group"].loc[frames["group"]["group"].str.contains(oldNode[0:2] + " ")] 
            oldGroups = pd.concat([oldGroups, oldGroups.rename(columns = {"group": "newGroup"})["newGroup"].str.replace(oldNode[0:2], newNode[0:2])], axis = 1)
            groupDict_df = pd.concat([groupDict_df, oldGroups])

    groupDict_df = groupDict_df.loc[groupDict_df["newGroup"] != groupDict_df["group"]].drop_duplicates(["group", "newGroup"])
    groupDict_df.set_index("group", inplace = True)

    frames["node"].index = frames["node"]["node"]
    frames["node"] = nodeDict_df["newNode"].combine_first(frames["node"]["node"]).drop_duplicates().to_frame(name="node")
    
    frames["unit"].index = frames["unit"]["unit"]
    frames["unit"] = unitDict_df["newUnit"].combine_first(frames["unit"]["unit"]).drop_duplicates().to_frame(name="unit")
    
    frames["effLevelGroupUnit"]["unit"] = frames["effLevelGroupUnit"]["unit"].replace(unitDict_df["newUnit"])
    frames["effLevelGroupUnit"] = frames["effLevelGroupUnit"].drop_duplicates(["effLevel", "unit"])
    
    frames["p_gn"]["node"] = frames["p_gn"]["node"].replace(nodeDict_df["newNode"])
    frames["p_gn"] = frames["p_gn"].drop_duplicates(["node"])

    frames["p_gnn"]["from node"] = frames["p_gnn"]["from node"].replace(nodeDict_df["newNode"])
    frames["p_gnn"]["to node"] = frames["p_gnn"]["to node"].replace(nodeDict_df["newNode"])
    frames["p_gnn"] = frames["p_gnn"].loc[frames["p_gnn"]["from node"] != frames["p_gnn"]["to node"]]
    frames["p_gnn"] = frames["p_gnn"].groupby(["grid", "from node", "to node"]).agg({"availability": "mean", "transferCap": "sum", "transferLoss": "mean", "invCost": "mean", "annuityFactor": "mean", "transferCapInvLimit": "sum"}).reset_index()
    
    frames["p_gnu_io"]["node"] = frames["p_gnu_io"]["node"].replace(nodeDict_df["newNode"])
    frames["p_gnu_io"]["unit"] = frames["p_gnu_io"]["unit"].replace(unitDict_df["newUnit"])
    frames["p_gnu_io"]["capacity"] = frames["p_gnu_io"]["capacity"].replace("eps", 0)
    frames["p_gnu_io"] = frames["p_gnu_io"].groupby(["grid", "node", "unit", "input output"]).agg({"conversionCoeff": "mean", "capacity": "sum", "unitSize": "mean", "vomCosts": "mean", "invCosts": "mean", "fomCosts": "mean", "annuityFactor": "mean",
                                                                                "upperLimitCapacityRatio": "mean"})
    frames["p_gnu_io"] = frames["p_gnu_io"].reset_index()    
    frames["p_gnu_io"].loc[frames["p_gnu_io"]["input output"] == "output", "capacity"] = frames["p_gnu_io"].loc[frames["p_gnu_io"]["input output"] == "output", "capacity"].replace(0, "eps")

    frames["p_unit"]["unit"] = frames["p_unit"]["unit"].replace(unitDict_df["newUnit"])
    frames["p_unit"]["eff00"] = frames["p_unit"]["eff00"].replace("", 0)
    frames["p_unit"]["maxUnitCount"] = frames["p_unit"]["maxUnitCount"].replace("inf", np.inf)
    frames["p_unit"] = frames["p_unit"].groupby("unit").agg({"eff00": "mean", "availability": "first", "maxUnitCount": "sum"}).reset_index()
    frames["p_unit"]["maxUnitCount"] = frames["p_unit"]["maxUnitCount"].replace(np.inf, "inf")

    frames["ts_influx"]["node"] = frames["ts_influx"]["node"].replace(nodeDict_df["newNode"])
    frames["ts_influx"] = frames["ts_influx"].groupby(["grid", "node", "forecast index"]).sum().reset_index()

    frames["ts_node"]["node"] = frames["ts_node"]["node"].replace(nodeDict_df["newNode"])
    frames["ts_node"] = frames["ts_node"].groupby(["grid", "node", "param_gnBoundaryTypes", "forecast index"]).sum().reset_index()
    
    frames["flowUnit"]["unit"] = frames["flowUnit"]["unit"].replace(unitDict_df["newUnit"])
    frames["flowUnit"] = frames["flowUnit"].drop_duplicates(["unit"])

    frames["ts_cf"]["node"] = frames["ts_cf"]["node"].replace(nodeDict_df["newNode"])
    frames["ts_cf"] = frames["ts_cf"].groupby(["flow", "node", "forecast index"]).mean().reset_index()

    frames["utAvailabilityLimits"]["unit"] = frames["utAvailabilityLimits"]["unit"].replace(unitDict_df["newUnit"])
    frames["utAvailabilityLimits"] = frames["utAvailabilityLimits"].drop_duplicates(["unit"])

    frames["p_gnBoundaryPropertiesForStates"]["node"] = frames["p_gnBoundaryPropertiesForStates"]["node"].replace(nodeDict_df["newNode"])
    frames["p_gnBoundaryPropertiesForStates"] = frames["p_gnBoundaryPropertiesForStates"].groupby(["grid", "node", "param_gnBoundaryTypes"]).agg({"useConstant": "first", "constant": "sum", "useTimeseries": "first"}).reset_index()
    
    frames["gnss_bound"]["node"] = frames["gnss_bound"]["node"].replace(nodeDict_df["newNode"])
    frames["gnss_bound"] = frames["gnss_bound"].drop_duplicates(["node"])

    frames["group"].index = frames["group"]["group"]
    frames["group"] = groupDict_df["newGroup"].combine_first(frames["group"]["group"]).drop_duplicates().to_frame(name="group")
    
    frames["ugroup"]["unit"] = frames["ugroup"]["unit"].replace(unitDict_df["newUnit"])
    frames["ugroup"]["group"] = frames["ugroup"]["group"].replace(groupDict_df["newGroup"])
    frames["ugroup"] = frames["ugroup"].drop_duplicates(["unit", "group"])

    frames["gnugroup"]["node"] = frames["gnugroup"]["node"].replace(nodeDict_df["newNode"])
    frames["gnugroup"]["unit"] = frames["gnugroup"]["unit"].replace(unitDict_df["newUnit"])
    frames["gnugroup"]["group"] = frames["gnugroup"]["group"].replace(groupDict_df["newGroup"])
    frames["gnugroup"] = frames["gnugroup"].drop_duplicates(["node", "unit", "group"])

    frames["gngroup"]["node"] = frames["gngroup"]["node"].replace(nodeDict_df["newNode"])
    frames["gngroup"]["group"] = frames["gngroup"]["group"].replace(groupDict_df["newGroup"])
    frames["gngroup"] = frames["gngroup"].drop_duplicates(["node", "group"])

    frames["sgroup"]["group"] = frames["sgroup"]["group"].replace(groupDict_df["newGroup"])
    frames["sgroup"] = frames["sgroup"].drop_duplicates(["group"])

    # does not work for energyShares!
    frames["p_groupPolicy"]["group"] = frames["p_groupPolicy"]["group"].replace(groupDict_df["newGroup"])
    frames["p_groupPolicy"] = frames["p_groupPolicy"].groupby(["group", "policy"]).sum().reset_index()

    frames["p_groupPolicyUnit"]["unit"] = frames["p_groupPolicyUnit"]["unit"].replace(unitDict_df["newUnit"])
    frames["p_groupPolicyUnit"]["group"] = frames["p_groupPolicyUnit"]["group"].replace(groupDict_df["newGroup"])
    frames["p_groupPolicyUnit"] = frames["p_groupPolicyUnit"].drop_duplicates(["group", "unit"])

    frames["p_groupPolicyEmission"]["group"] = frames["p_groupPolicyEmission"]["group"].replace(groupDict_df["newGroup"])
    frames["p_groupPolicyEmission"] = frames["p_groupPolicyEmission"].drop_duplicates(["group"])

    europe2Letter = ['AL', 'AT', 'BA', 'BE', 'BG', 'CH', 'CZ', 'DE', 'DK', 'EE', 'ES', 'FI', 'FR', 'GB', 'GR', 'HR', 'HU',
                 'IE', 'IT', 'LT', 'LU', 'LV', 'ME', 'MK', 'NL', 'NO', 'PL', 'PT', 'RO', 'RS', 'SE', 'SI', 'SK']
    europe3Letter = ['ALB', 'AUT', 'BIH', 'BEL', 'BGR', 'CHE', 'CZE', 'DEU', 'DNK', 'EST', 'ESP', 'FIN', 'FRA', 'GBR', 'GRC',
                    'HRV', 'HUN', 'IRL', 'ITA', 'LTU', 'LUX', 'LVA', 'MNE', 'MKD', 'NLD', 'NOR', 'POL', 'PRT', 'ROU', 'SRB',
                    'SWE', 'SVN', 'SVK']

    twoToThree = dict(zip(europe2Letter, europe3Letter))
    
    for path in [f'pypsa-eur/resources/regions_onshore_elec_s_{snakemake.wildcards.clusters}.geojson', 
                 f'pypsa-eur/resources/regions_offshore_elec_s_{snakemake.wildcards.clusters}.geojson']:
        map = gpd.read_file(path).set_index("name")
        map["newNode"] = map.index
        map["countryCode"] = map.index.str[0:2]
        map["countryCode"] = map["countryCode"].replace(twoToThree)
        for newNode in nodesToAggregate.keys():
            for oldNode in nodesToAggregate[newNode]:
                map.loc[oldNode, "newNode"] = newNode
        map = map.dissolve(by = "newNode")
        for name in map.index:
            if name in nodesToAggregate.keys():
                countryList = nodesToAggregate[name]
                countryCodeList = list(set([twoToThree[country[0:2]] for country in countryList]))
                if len(countryCodeList) == 1:
                    countryCodeList = countryCodeList[0]
                else:
                    countryCodeList = ';'.join(countryCodeList)
                map.at[name, "countryCode"] = countryCodeList
        map = map.reset_index().rename(columns={"newNode": "name"})
        if "onshore" in path:
            map.to_file(os.path.join('results', f'regions_onshore_elec_s_{snakemake.wildcards.clusters}_agg.geojson'))
        else:
            map.to_file(os.path.join('results', f'regions_offshore_elec_s_{snakemake.wildcards.clusters}_agg.geojson'))

if not os.path.isdir(os.path.split(config["pypsa2bbPaths"]["output_path"])[0]):
    os.mkdir(os.path.split(config["pypsa2bbPaths"]["output_path"])[0])

with pd.ExcelWriter(config["pypsa2bbPaths"]["output_path"]) as writer:
    for key in frames:
        frames[key].to_excel(writer, sheet_name=key, index=False)