'''
    pypsa2bb - create input data for the esm backbone, run backbone and analyse results
    Copyright (C) 2020-2022 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from functools import partial
import numpy as np
from gams import GamsWorkspace, GamsException
import pandas as pd
from pathlib import Path
import re
import subprocess as sp
from io import StringIO
from matplotlib.pyplot import get_cmap
from matplotlib.colors import to_hex
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots

class PypsaResult:
    def __init__(self, path):
        import pypsa

        self.network = pypsa.Network(path)
        self.country_frames = self._get_country_frames()

    @property
    def carriers(self) -> list:
        realized_generation = self.network.generators_t.p
        realized_storage = self.network.storage_units_t.p
        realized_production = pd.concat([realized_generation, realized_storage], axis=0)
        # extract e.g. "solar" from "DE0 0 solar"
        return [x.rsplit(" ", 1)[-1] for x in realized_production.columns]

    @property
    def countries(self) -> list:
        realized_generation = self.network.generators_t.p
        realized_storage = self.network.storage_units_t.p
        realized_production = pd.concat([realized_generation, realized_storage], axis=0)
        # extract e.g. "DE" from "DE0 0 solar"
        return [x[:2] for x in realized_production.columns]

    def _get_country_frames(self, skip_first=True) -> dict:
        realized_generation = self.network.generators_t.p.T
        realized_storage = self.network.storage_units_t.p.T

        realized_production = pd.concat([realized_generation, realized_storage], axis=0)

        realized_production.insert(0, "carrier", self.carriers, False)
        realized_production.insert(0, "country", self.countries, False)

        # select a subset of the DataFrame by country
        country_ids = set(self.countries)
        pypsa_country_dfs = [
            realized_production[realized_production["country"] == c]
            for c in country_ids
        ]
        pypsa_country_dfs = [x.groupby("carrier").sum(numeric_only=True).T for x in pypsa_country_dfs]

        # pypsa_country_dfs[0]["PHS"] *=-0.866025

        if skip_first:
            # skip first row, since it is not present in backbone
            pypsa_country_dfs = [x.iloc[1:, :] for x in pypsa_country_dfs]
        return dict(zip(country_ids, pypsa_country_dfs))


class BackboneResult:
    def __init__(self, path):
        ws = GamsWorkspace()
        self._path = Path(path).absolute().as_posix()
        try:
            self.gams_db = ws.add_database_from_gdx(self._path)
            self.gams_db = ws.add_database_from_gdx(self._path)
        except GamsException as ge:
            raise GamsException(f"Error when opening {self._path}").with_traceback(ge.__traceback__)
        except TypeError as te:
            raise TypeError(f"An error occured on {self._path}:").with_traceback(te.__traceback__)

        self.mMettings = self.gams_db.get_parameter("r_info_mSettings")

        # invest or schedule
        self.model_type = self.mMettings.first_record().keys[0]
        self.stepLengthInHours = self.mMettings.find_record(
            keys=["stepLengthInHours"]
        ).value

        # allows easy access of results
        self.set_results_as_attribs()
        self.has_dummy_generation = not self.r_qGen_g().empty
        if self.has_dummy_generation:
            print(f"WARNING: Model has dummy generation. ({Path(path).name})")

    def set_results_as_attribs(self):
        for symbol in self._result_symbols:
            # this allows to access result parameters by BackboneResult.r_emission()
            setattr(self, symbol, partial(self.param_as_df_gdxdump, symbol))

    @property
    def _backbone_path(self):
        """path to the backbone repository
        """
        return Path(__file__).parent.parent.joinpath("backbone")

    @property
    def _result_symbols(self):
        """list of possible symbols in the {result_file}.gdx as defined in backbone/defOutput/resultSymbols.inc
        """
        result_symbole_file = self._backbone_path.joinpath(
            "defOutput/resultSymbols.inc"
        )
        with result_symbole_file.open("r") as symbol_file:
            text = symbol_file.read()
            symbols = re.findall(r"^\br_[^\s]+\b$", text, re.MULTILINE)
        return symbols

    def param_as_df(self, param_name, convert_time=True):
        gms_param = self.gams_db.get_parameter(param_name)
        data = [[*rec.keys, rec.value] for rec in gms_param]

        # domains have changed and now may contain sets
        domains = [x if type(x) == str else x.name for x in gms_param.domains]

        columns = domains + ["Val"]
        param_df = pd.DataFrame(data=data, columns=columns)

        # converts the time column to int by omitting the first character
        if convert_time:
            if "t" in param_df.columns:
                param_df["t"] = param_df["t"].apply(lambda t: int(t[1:]))

        return param_df

    def param_as_df_gdxdump(self, param_name, encoding="1252", convert_time=True):
        """ Use the 'gdxdump' GAMS utility via subprocess to convert a parameter into a pd.DataFrame.
        This is sometimes beneficial, to circumvent decoding errors
        """
        gdxdump = sp.run(
            ["gdxdump", self._path, "format", "csv", "symb", param_name], stdout=sp.PIPE
        )
        csv_data = gdxdump.stdout.decode(encoding=encoding)
        header = csv_data.partition("\n")[0]

        header = [x.strip('"') for x in header.split(",")]
        dtypes = [str if x != "Val" else float for x in header]
        dtypes = dict(zip(header, dtypes))
        df = pd.read_csv(StringIO(csv_data), dtype=dtypes, na_values="Eps")

        # converts the time column to int by omitting the first character
        if convert_time:
            if "t" in df.columns:
                df["t"] = df["t"].apply(lambda t: int(t[1:]))
        return df

    def eqn_as_df(self, eqn_name):
        gms_eqn = self.gams_db.get_equation(eqn_name)

        data = [
            [*rec.keys, rec.level, rec.marginal, rec.lower, rec.upper, rec.scale]
            for rec in gms_eqn
        ]
        columns = [x if type(x) == str else x.name for x in gms_eqn.domains] + [
            "level",
            "marginal",
            "lower",
            "upper",
            "scale",
        ]

        eqn_df = pd.DataFrame(data=data, columns=columns)
        return eqn_df

    def set_as_df(self, set_name):
        gms_set = self.gams_db.get_set(set_name)
        data = [[*rec.keys] for rec in gms_set]
        domains = [x if type(x) == str else x.name for x in gms_set.domains]

        set_df = pd.DataFrame(data=data,columns=domains)
        return set_df

    @property
    def generation_frame(self):

        # avoid unneccessary calculation and loading of data
        if hasattr(self, "_gen_df"):
            return self._gen_df

        bb_gen_utype = self.param_as_df("r_genByUnittype_gnft")
        countries = pd.DataFrame(bb_gen_utype["node"].str[:2])
        bb_gen_utype["country"] = countries

        for ix, element in enumerate(bb_gen_utype["grid"]):
            if element == "storage" and bb_gen_utype["unittype"][ix] == "PHS":
                bb_gen_utype["Val"][ix] = bb_gen_utype["Val"][ix] * (-0.866025)

        # transformation from long to wide dataformat
        gen_df = bb_gen_utype.pivot_table(
            columns=["country", "unittype"], values=["Val"], index="t", aggfunc=sum
        )
        # get rid of "Val"-level
        gen_df.columns = gen_df.columns.droplevel()
        self._gen_df = gen_df
        return self._gen_df

    @property
    def carriers(self):
        return self.generation_frame.columns.get_level_values(1)

    def plot_load_profile(self, country="DE"):
        ax = self.gen_df_safe[country].plot(
            kind="area", figsize=(15, 8), colormap="tab20"
        )
        ax.set_ylabel("MW")
        return ax

    def plot_generation_sum_pie(self, country="DE"):
        agg_supply = self.aggregated_per_carrier_supply(safe=True)
        ax = agg_supply.plot(kind="pie", figsize=(15, 8), autopct="%.2f")
        total_sum = agg_supply.sum() * self.stepLengthInHours
        ax.set_ylabel(f"Share of total generation ({total_sum/10**6:.2f} TWh) [%]")
        return ax

    def aggregated_per_carrier_supply(self, country="DE", percent=False, safe=False):
        if safe:
            agg_supply = self.gen_df_safe[country].sum()
        else:
            agg_supply = self.generation_frame[country].sum()
        if percent:
            return (agg_supply / agg_supply.sum() * 100).apply(lambda x: f"{x:.2f}")
        else:
            return agg_supply

    @property
    def gen_df_safe(self):
        """
        returns a pd.DataFrame that is save to use for plotting. This is required, 
        since negative values can occur in the solution.
        """
        if hasattr(self, "_gen_df_safe"):
            return self._gen_df_safe

        # generate df with True/False values
        negative_vals = self.generation_frame < 0

        # sum for better display
        per_carrier_neg_count = negative_vals.sum()
        if not all(per_carrier_neg_count == 0):
            print(
                f"!!! negative values in generation frame. Listing negative counts !!!\n{per_carrier_neg_count}"
            )
            neg_cols = per_carrier_neg_count != 0
            neg_cols = [idx for idx, val in per_carrier_neg_count.items() if val != 0]

            neg_vals_times = [
                self.generation_frame.loc[negative_vals[idx], idx] for idx in neg_cols
            ]
            print(f"Detail:\n{neg_vals_times}")

            # set property to avoid unnesseccary calls
            self._gen_df_safe = abs(self.generation_frame)
        else:
            self._gen_df_safe = self.generation_frame

        return self._gen_df_safe

    def plot_load_duration(self, country="DE"):
        ax = self.generation_frame.sum(axis=1).sort_values(ascending=False).plot()
        ax.set_ylabel("MW")
        return ax

    def get_country_frames(self):

        bb_gen = self.generation_frame
        # bb_gen["DE","coal"]
        # level 0 is country, level 1 is unittype
        country_ids = set(bb_gen.columns.get_level_values(0))

        bb_country_dfs = [bb_gen[c] for c in country_ids]
        # replace nans with zeros
        bb_country_dfs = [df.fillna(0) for df in bb_country_dfs]
        return dict(zip(country_ids, bb_country_dfs))


class BackboneScenarioAnalysis:
    def __init__(self, results_dir=None, result_files=None) -> None:
        if results_dir and not result_files:
            abs_dir = Path(results_dir).resolve()
            unsorted_results = list(abs_dir.glob(r"*.gdx"))
            self.result_files = sorted(unsorted_results)
        elif not results_dir and result_files:
            self.result_files = sorted([Path(path) for path in result_files])
        else:
            raise ValueError(
                f"""One of the parameters 'results_dir' or 'result_files' must be passed.
                Received: {dict(results_dir=results_dir,result_files=result_files)}.
                """
            )

        self.bb_results = [
            BackboneResult(Path(file).as_posix()) for file in self.result_files
        ]

        # scenario names will be derived from the filenames
        self.scenarios = [Path(file).stem.replace("_", " ") for file in self.result_files]

        # define colors as sequence:
        self.colors = self._get_n_hex_colors_from_cmap(
            "Spectral_r", len(self.result_files)
        )

        # define colors also as sequence for the number of scenarios: 
        self.colors_scen = self._get_n_hex_colors_from_cmap(
            "Spectral_r", len(self.scenarios)
        )
        self.set_results_as_attribs()
        pass


    def _scenario_param(self, param_name):
        frames = []
        for result, scenario in zip(self.bb_results, self.scenarios):
            frame = getattr(result, param_name)()
            frame["Scenario"] = scenario
            frames.append(frame)
        return pd.concat(frames)

    def set_results_as_attribs(self):
        for symbol in self.bb_results[0]._result_symbols:
            # this allows to access result parameters by BackboneResult.r_emission()
            setattr(self, symbol, partial(self._scenario_param, symbol))

    @staticmethod
    def _get_n_hex_colors_from_cmap(cmap_name, n):
        cmap = get_cmap(cmap_name)
        fracs = np.linspace(0, 1, n)
        return [to_hex(cmap(f)) for f in fracs]

    def installed_capacity(self) -> pd.DataFrame:

        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            installed_capacity = result.r_invest_unitCapacity_gnu().groupby("unit").sum(numeric_only=True)

            installed_capacity["scenario"] = scen

            frames.append(installed_capacity)

        scenario_df = pd.concat(frames)
        scenario_df = scenario_df.reset_index()
        return scenario_df

    def plot_installed_capacity(
        self, colors=None
    ):
        if colors == None:
            colors = self.colors
        scenario_df = self.installed_capacity()
        fig = px.bar(
            scenario_df,
            x="unit",
            y="Val",
            color="scenario",
            barmode="group",
            color_discrete_sequence=colors,
        )

        return fig

    def system_cost_composition(self,) -> pd.DataFrame:

        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            # print(scen, result._path)
            invest_cost = result.r_cost_unitInvestmentCost_gnu()
            invest_cost["Category"] = "Invest"
            invest_cost_sum = invest_cost.groupby("Category").sum(numeric_only=True)
            
            transmission_cost = result.r_cost_linkInvestmentCost_gnn()
            transmission_cost["Category"] = "Grid expansion"
            transmission_cost_sum = transmission_cost.groupby("Category").sum(numeric_only=True)

            fuel_emission_cost = result.r_cost_unitFuelEmissionCost_u()
            fuel_emission_cost["Category"] = fuel_emission_cost["node"].str.capitalize()
            fuel_emission_cost_sum = fuel_emission_cost.groupby("Category").sum(numeric_only=True)

            fom_cost = result.r_cost_unitFOMCost_gnu()
            fom_cost["Category"] = "FOM"
            fom_cost_sum = fom_cost.groupby("Category").sum(numeric_only=True)

            vom_cost = result.r_cost_unitVOMCost_gnu()
            vom_cost["Category"] = "VOM"
            vom_cost_sum = vom_cost.groupby("Category").sum(numeric_only=True)

            cost_df = pd.concat([invest_cost_sum, fuel_emission_cost_sum, fom_cost_sum, vom_cost_sum, transmission_cost_sum])
            cost_df["Scenario"] = scen
            frames.append(cost_df)
        scenario_df = pd.concat(frames)
        scenario_df = scenario_df.reset_index()
        return scenario_df

    def plot_system_cost_composition(self):
        scenario_df = self.system_cost_composition()
        fig = px.bar(scenario_df, x="Scenario", y="Val", color="Category", width=700)
        fig.update_yaxes({"title": "annual system cost [M€]"})
        fig.update_layout(legend={'traceorder': 'reversed'})
        # fig.update_traces(dict(width=[0.6] * 8))
        return fig

    def energy_provided_per_unitType(
        self, unitType_mapping: dict = None
    ) -> pd.DataFrame:
        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            sample_df = result.r_gen_gnu()
            grouped_df = sample_df.groupby(["grid", "unit"]).sum(numeric_only=True)
            grouped_df["scenario"] = scen
            frames.append(grouped_df)
        scenario_df = pd.concat(frames)
        return scenario_df

    def plot_energy_per_unitType(self, colors=None):

        if colors is None:
            colors = self.colors
        fig = go.Figure()

        scenario_df = self.energy_provided_per_unitType().reset_index()

        for i, scen in enumerate(self.scenarios):
            data = scenario_df.loc[scenario_df["scenario"] == scen, :]
            x = []
            x.append(data["grid"].values)
            x.append(data["unit"].values)
            color = colors[i % len(colors)]
            fig.add_bar(x=x, y=data["Val"], name=scen, marker={"color": color})
        # fig.update_traces(colors = px.colors.sequential.RdBu[:4])
        fig.update_yaxes(title="provided energy by unit type [MWh]")
        # fig.write_image(f"{results_dir}/figures/energy_generation_by_type.svg")
        # fig.write_image(f"{results_dir}/figures/energy_generation_by_type.pdf")
        return fig

    def plot_energy_per_unitType_stacked(self, colors=None, u_name_filter=None):
        fig = go.Figure()
        scenario_df = self.energy_provided_per_unitType().reset_index()
        
        if u_name_filter:
            # for a birds eye view, this can be helpful for condensing unit_names 
            # i.e. u_name_filter=lambda x: x.split(" ",2)[-1]
            # "DE0 0 coal" -> coal 
            # "FR0 1 coal" -> coal
            scenario_df["unit"] = scenario_df["unit"].apply(u_name_filter)

        utypes = scenario_df["unit"].unique()


        if colors is None:
            colors = self._get_n_hex_colors_from_cmap(
            "tab20", len(utypes)
        )

        for i, utype in enumerate(utypes):

            # select and aggregate data
            data = scenario_df.loc[scenario_df["unit"] == utype, :]
            data = data.groupby(["grid","unit","scenario"]).sum().reset_index()
            
            # generate double index
            x = []
            x.append(data["grid"].values)
            x.append(data["scenario"].values)
            color = colors[i % len(colors)]
            fig.add_bar(x=x, y=data["Val"], name=utype, marker={"color": color})
        fig.update_yaxes(title="provided energy by unit type [MWh]")
        fig.update_layout(barmode="relative")
        return fig

    def emissions_avoidance(self):
        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            # print(scen, result._path)
            emissions = result.r_emission_n()
            emissions["scenario"] = scen
            frames.append(emissions)

        emissions_df = pd.concat(frames)
        unique_emissions = emissions_df.emission.unique()
        assert len(unique_emissions) == 1, NotImplementedError(
            f"Multiple emissions not yet implemented. Found {unique_emissions}"
        )

        total_emissions_df = emissions_df.groupby("scenario").sum(numeric_only=True)
        avoided_emissions = (
            total_emissions_df["Val"].max() - total_emissions_df["Val"]
        )

        cost_df = self.system_cost_composition()
        agg_scenario_cost_df = cost_df.groupby("scenario").sum(numeric_only=True)["Val"]
        cost_for_avoidance = (
            agg_scenario_cost_df / avoided_emissions
        )
        cost_for_avoidance = cost_for_avoidance * 10 ** 6
        avoidance_cost_df = cost_for_avoidance.to_frame(name="Val").reset_index()

        fig = make_subplots(specs=[[{"secondary_y": True}]])
        for node in emissions_df.node.unique():
            trace = go.Bar(
                x=emissions_df.loc[emissions_df.node == node, "scenario"],
                y=emissions_df.loc[emissions_df.node == node, "Val"],
                name=node,
            )
            fig.add_trace(trace)

        trace = go.Scatter(
            x=avoidance_cost_df["scenario"],
            y=avoidance_cost_df["Val"],
            name="CO2 avoidance cost",
            mode="markers",
            marker=dict(color="black"),
        )
        fig.add_trace(trace, secondary_y=True)
        fig.update_layout(barmode="relative")
        fig.update_layout(
            yaxis=dict(
                title=r"$Emissions\ [t_{CO_2}]$",
                titlefont=dict(color="black"),
                tickfont=dict(color="black"),
                # family="sans-serif"
            ),
            yaxis2=dict(
                title=r"$Cost\ [\frac{\text{€}}{t_{CO_2}}]$",
                titlefont=dict(color="black", family="sans-serif"),
                tickfont=dict(color="black"),
            ),
            legend=dict(
                x=0.95,
                y=0.95,
                traceorder="normal",
                font=dict(family="sans-serif", size=12, color="black"),
                bordercolor="darkgray",
                borderwidth=2,
                xanchor="right",
            ),
        )
        # fig.write_image(
        #     "inputs/ses2021/multi_building_model/figures/final_emissions_avoidance_cost.svg",
        #     format="svg",
        # )
        return fig

    # Plot Price Duration Curve by scenarios over all or predefined nodes in the electric grid
    def plot_priceDurationCurve(self, nodes=[],stylelist=[],colorlist=[]):
        result = self.bb_results[0]
        marginal = result.r_balance_marginalValue_gnft()
        # extract all electric nodes from first scenario if non are given
        # grid choice has to be adapted or dropped in the code if desired
        if not nodes:
            nodes = marginal[marginal['grid'] == 'elec']['node'].drop_duplicates()
        ax_list = []
        # define list for graph's line style
        if not stylelist:
            for x in range(len(self.scenarios)):
                for element in  ['-', '--', '-.', ':']:
                    stylelist.append(element)
        if not colorlist:
            colorlist=self.colors_scen
        # create plots by node over scenarios
        for n in nodes:
            scen_dict = {}
            for s, r in zip(self.scenarios, self.bb_results):
                marginal = r.r_balance_marginalValue_gnft()
                marginal = marginal[marginal['node'] == str(n)].sort_values('Val')
                marginal['Val'] = marginal['Val']*-1
                price = marginal['Val'].tolist()
                scen_dict.update({str(s): price})
                t = [np.arange(0, len(price), 1)]
            df = pd.DataFrame(scen_dict, index=t)
            ax = df.plot.line(title=str(n), lw=2, style=stylelist,color=colorlist)
            ax.set_xticks([])
            ax.set_yticks(range(min(0, int(min(price))), int(max(price)),
                          round((int(max(price))-min(0, int(min(price))))/10)))
            ax.set_xlabel('hours sorted')
            ax.set_ylabel('€/MWh')
            ax.legend()
            ax_list.append(ax)
        return ax_list

    # Plot Price Duration Curve by all or predefined electric nodes over scenario
    def plot_priceDurationCurve2(self,nodes=[],stylelist=[],colorlist=[]):
        result = self.bb_results[0]
        marginal = result.r_balance_marginalValue_gnft()
        # extract all electric nodes from first scenario if non are given
        # grid choice has to be adapted or dropped in the code if desired
        if not nodes:
            nodes = marginal[marginal['grid'] == "elec"]["node"].drop_duplicates().transpose().tolist()
        scen_dict = {}
        ax_list = []
        # define list for graph's line style
        if not stylelist:
            for x in range(len(nodes)):
                for element in  ['-', '--', '-.', ':']:
                    stylelist.append(element)
        if not colorlist:
            colorlist=self._get_n_hex_colors_from_cmap("Spectral_r",len(nodes))
        # create plots by nodes over scenarios
        for s, r in zip(self.scenarios, self.bb_results):
            for n in nodes:
                # result=self.bb_results[0]
                marginal = r.r_balance_marginalValue_gnft()
                marginal = marginal[marginal['node'] == str(n)].sort_values('Val')
                marginal["Val"] = marginal["Val"]*-1
                price = marginal["Val"].tolist()
                t = [np.arange(0, len(price), 1)]
                scen_dict.update({str(n): price})
                df = pd.DataFrame(scen_dict, index=t)
            ax = df.plot.line(title=str(s),style=stylelist,lw=2,color=colorlist)
            ax.set_xticks([])
            ax.set_yticks(range(min(0, int(min(price))), int(max(price)),
                          round((int(max(price))-min(0, int(min(price))))/10)))
            ax.set_xlabel('hours sorted')
            ax.set_ylabel('€/MWh')
            ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
            ax_list.append(ax)
        return ax_list


if __name__ == "__main__":
    result_files = Path("results").glob("*result.gdx")
    scen_lysis = BackboneScenarioAnalysis(result_files=result_files)

    stacked_generation = scen_lysis.plot_energy_per_unitType_stacked(u_name_filter=lambda x: x.split(" ",2)[-1])
    pass
